from __future__ import annotations

from typing import Dict, List, Optional, Tuple

import netsquid as ns
import pytest
from netqasm.lang.instr.flavour import (
    Flavour,
    NVFlavour,
    TrappedIonFlavour,
    VanillaFlavour,
)
from netqasm.lang.parsing import parse_text_subroutine
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.nodes import Node
from netsquid.qubits import ketstates

from qoala.lang.ehi import UnitModule
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.routine import LocalRoutine, RoutineMetadata
from qoala.runtime.config import NvParams, TopologyConfig
from qoala.runtime.lhi import LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.ntf import GenericNtf, NvNtf, TrappedIonNtf
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.memmgr import AllocError, MemoryManager, NotAllocatedError
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice
from qoala.sim.qnos import (
    GenericProcessor,
    NVProcessor,
    QnosComponent,
    QnosInterface,
    QnosLatencies,
    QnosProcessor,
)
from qoala.sim.qnos.qnosprocessor import (
    IonTrapProcessor,
    UnsupportedNetqasmInstructionError,
)
from qoala.util.math import has_multi_state, has_state
from qoala.util.tests import netsquid_run


def perfect_uniform_qdevice(num_qubits: int) -> QDevice:
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits=num_qubits,
        single_instructions=[
            INSTR_INIT,
            INSTR_X,
            INSTR_Y,
            INSTR_Z,
            INSTR_H,
            INSTR_ROT_X,
            INSTR_ROT_Y,
            INSTR_ROT_Z,
            INSTR_MEASURE,
        ],
        single_duration=5e3,
        two_instructions=[INSTR_CNOT],
        two_duration=100e3,
    )
    processor = build_qprocessor_from_topology(name="processor", topology=topology)
    node = Node(name="alice", qmemory=processor)
    return QDevice(node=node, topology=topology)


def perfect_nv_star_qdevice(num_qubits: int) -> QDevice:
    cfg = TopologyConfig.from_nv_params(num_qubits, NvParams())
    topology = LhiTopologyBuilder.from_config(cfg)
    processor = build_qprocessor_from_topology(name="processor", topology=topology)
    node = Node(name="alice", qmemory=processor)
    return QDevice(node=node, topology=topology)


def perfect_trapped_ion_qdevice(num_qubits: int) -> QDevice:
    topology = LhiTopologyBuilder.trapped_ion_default_perfect_gates(num_qubits)
    processor = build_qprocessor_from_topology(name="processor", topology=topology)
    node = Node(name="alice", qmemory=processor)
    return QDevice(node=node, topology=topology)


def create_program(
    subroutines: Optional[Dict[str, LocalRoutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> QoalaProgram:
    if subroutines is None:
        subroutines = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    return QoalaProgram(blocks=[], local_routines=subroutines, meta=meta)


def create_process(
    pid: int, program: QoalaProgram, unit_module: UnitModule
) -> QoalaProcess:
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
    )
    mem = ProgramMemory(pid=pid)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def create_process_with_subrt(
    pid: int,
    subrt_text: str,
    unit_module: UnitModule,
    flavour: Flavour,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> QoalaProcess:
    if uses is None:
        uses = []
    if keeps is None:
        keeps = []
    subrt = parse_text_subroutine(subrt_text, flavour=flavour)
    metadata = RoutineMetadata(uses, keeps)
    iqoala_subrt = LocalRoutine("subrt", subrt, return_vars=[], metadata=metadata)
    meta = ProgramMeta.empty("alice")
    meta.epr_sockets = {0: "bob"}
    program = create_program(subroutines={"subrt": iqoala_subrt}, meta=meta)
    return create_process(pid, program, unit_module)


def create_process_with_vanilla_subrt(
    pid: int,
    subrt_text: str,
    unit_module: UnitModule,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> QoalaProcess:
    return create_process_with_subrt(
        pid, subrt_text, unit_module, VanillaFlavour(), uses, keeps
    )


def create_process_with_nv_subrt(
    pid: int,
    subrt_text: str,
    unit_module: UnitModule,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> QoalaProcess:
    return create_process_with_subrt(
        pid, subrt_text, unit_module, NVFlavour(), uses, keeps
    )


def create_process_with_trapped_ion_subrt(
    pid: int,
    subrt_text: str,
    unit_module: UnitModule,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> QoalaProcess:
    return create_process_with_subrt(
        pid, subrt_text, unit_module, TrappedIonFlavour(), uses, keeps
    )


def set_new_subroutine(
    process: QoalaProcess,
    subrt_text: str,
    flavour: Flavour,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> None:
    if uses is None:
        uses = []
    if keeps is None:
        keeps = []
    subrt = parse_text_subroutine(subrt_text, flavour=flavour)
    metadata = RoutineMetadata(uses, keeps)
    iqoala_subrt = LocalRoutine("subrt", subrt, return_vars=[], metadata=metadata)
    program = process.prog_instance.program
    program.local_routines["subrt"] = iqoala_subrt


def set_new_vanilla_subroutine(
    process: QoalaProcess,
    subrt_text: str,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> None:
    set_new_subroutine(process, subrt_text, VanillaFlavour(), uses, keeps)


def set_new_nv_subroutine(
    process: QoalaProcess,
    subrt_text: str,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> None:
    set_new_subroutine(process, subrt_text, NVFlavour(), uses, keeps)


def set_new_trapped_ion_subroutine(
    process: QoalaProcess,
    subrt_text: str,
    uses: Optional[List[int]] = None,
    keeps: Optional[List[int]] = None,
) -> None:
    set_new_subroutine(process, subrt_text, TrappedIonFlavour(), uses, keeps)


def execute_process(processor: GenericProcessor, process: QoalaProcess) -> int:
    all_routines = process.program.local_routines
    routine = all_routines["subrt"]
    for virt_id in routine.metadata.qubit_use:
        if processor.interface.memmgr.phys_id_for(process.pid, virt_id) is None:
            processor.interface.memmgr.allocate(process.pid, virt_id)

    # input/result arrays not used
    # TODO: add tests that do use these
    inputs = process.inputs.values
    processor.instantiate_routine(process, routine, inputs, MemAddr(0), MemAddr(0))

    netqasm_instructions = routine.subroutine.instructions

    instr_count = 0

    instr_idx = 0
    while instr_idx < len(netqasm_instructions):
        instr_count += 1
        instr_idx = netsquid_run(
            processor.assign_routine_instr(process, "subrt", instr_idx)
        )
    return instr_count


def execute_multiple_processes(
    processor: GenericProcessor, processes: List[QoalaProcess]
) -> None:
    for proc in processes:
        all_routines = proc.program.local_routines
        routine = all_routines["subrt"]
        for virt_id in routine.metadata.qubit_use:
            if processor.interface.memmgr.phys_id_for(proc.pid, virt_id) is None:
                processor.interface.memmgr.allocate(proc.pid, virt_id)

        # input/result arrays not used
        # TODO: add tests that do use these
        inputs = proc.inputs.values
        processor.instantiate_routine(proc, routine, inputs, MemAddr(0), MemAddr(0))
        netqasm_instructions = routine.subroutine.instructions
        for i in range(len(netqasm_instructions)):
            netsquid_run(processor.assign_routine_instr(proc, "subrt", i))


def setup_components_generic(
    num_qubits: int, latencies: QnosLatencies = QnosLatencies.all_zero()
) -> Tuple[QnosProcessor, UnitModule]:
    qdevice = perfect_uniform_qdevice(num_qubits)
    ehi = LhiConverter.to_ehi(qdevice.topology, ntf=GenericNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    qnos_comp = QnosComponent(node=qdevice._node)
    memmgr = MemoryManager(qdevice._node.name, qdevice)
    interface = QnosInterface(qnos_comp, qdevice, memmgr)
    processor = GenericProcessor(interface, latencies)
    return processor, unit_module


def setup_components_nv_star(
    num_qubits: int, latencies: QnosLatencies = QnosLatencies.all_zero()
) -> Tuple[QnosProcessor, UnitModule]:
    qdevice = perfect_nv_star_qdevice(num_qubits)
    ehi = LhiConverter.to_ehi(qdevice.topology, ntf=NvNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    qnos_comp = QnosComponent(node=qdevice._node)
    memmgr = MemoryManager(qdevice._node.name, qdevice)
    interface = QnosInterface(qnos_comp, qdevice, memmgr)
    processor = NVProcessor(interface, latencies)
    return processor, unit_module


def setup_components_trapped_ion(
    num_qubits: int, latencies: QnosLatencies = QnosLatencies.all_zero()
) -> Tuple[QnosProcessor, UnitModule]:
    qdevice = perfect_trapped_ion_qdevice(num_qubits)
    ehi = LhiConverter.to_ehi(qdevice.topology, ntf=TrappedIonNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    qnos_comp = QnosComponent(node=qdevice._node)
    memmgr = MemoryManager(qdevice._node.name, qdevice)
    interface = QnosInterface(qnos_comp, qdevice, memmgr)
    processor = IonTrapProcessor(interface, latencies)
    return processor, unit_module


def test_init_qubit():
    num_qubits = 3
    processor, unit_module = setup_components_generic(num_qubits)

    subrt = """
    set Q0 0
    init Q0
    """

    process = create_process_with_vanilla_subrt(0, subrt, unit_module, [0], [0])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 0 has been initialized.
    phys_id = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)


def test_init_qubit_with_latencies():
    ns.sim_reset()

    instr_time = 1e3

    # TODO: improve this
    # Value is copied from hardcoded implementation of `perfect_uniform_qdevice`.
    gate_time = 5e3

    num_qubits = 3
    processor, unit_module = setup_components_generic(
        num_qubits, latencies=QnosLatencies(qnos_instr_time=instr_time)
    )

    subrt = """
    set Q0 0
    init Q0
    """

    process = create_process_with_vanilla_subrt(0, subrt, unit_module, [0], [0])
    processor._interface.memmgr.add_process(process)

    assert ns.sim_time() == 0
    execute_process(processor, process)
    assert ns.sim_time() == instr_time + 1 * gate_time

    # Check if qubit with virt ID 0 has been initialized.
    phys_id = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)


def test_init_not_allocated():
    num_qubits = 3
    processor, unit_module = setup_components_generic(num_qubits)

    subrt = """
    set Q0 0
    init Q0
    """

    # set "uses" to empty list so virt ID 0 is not allocated
    process = create_process_with_vanilla_subrt(0, subrt, unit_module, uses=[])
    processor._interface.memmgr.add_process(process)

    with pytest.raises(NotAllocatedError):
        execute_process(processor, process)


def test_single_gates_generic():
    num_qubits = 3
    processor, unit_module = setup_components_generic(num_qubits)

    subrt = """
    set Q0 0
    init Q0
    x Q0
    """

    process = create_process_with_vanilla_subrt(0, subrt, unit_module, [0], [0])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 0 has been initialized.
    phys_id = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)

    # Qubit should be in |1>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s1)

    # New subroutine: apply X.
    subrt = """
    x Q0
    """
    set_new_vanilla_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should be back in |0>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)

    # New subroutine: apply Z.
    subrt = """
    z Q0
    """
    set_new_vanilla_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should still be in |0>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)

    # New subroutine: apply PI/2 Y-rotation.
    # pi/2 = 8 / 2^4 * pi
    subrt = """
    rot_y Q0 8 4
    """
    set_new_vanilla_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should be in |+>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.h0)

    # New subroutine: apply -PI/2 Z-rotation.
    # -pi/2 = 24 / 2^4 * pi
    subrt = """
    rot_z Q0 24 4
    """
    set_new_vanilla_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should be in |-i>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.y1)

    # New subroutine: apply S on qubit 0 (comm). Should not be allowed.
    subrt = """
    s Q0
    """
    set_new_vanilla_subroutine(process, subrt, [0], [0])
    with pytest.raises(UnsupportedNetqasmInstructionError):
        execute_process(processor, process)


def test_single_gates_multiple_qubits_generic():
    num_qubits = 3
    processor, unit_module = setup_components_generic(num_qubits)

    # Initialize q0 and q1. Apply X on q0 and Z on q1.
    subrt = """
    set Q0 0
    init Q0
    set Q1 1
    init Q1
    x Q0
    z Q1
    """

    process = create_process_with_vanilla_subrt(0, subrt, unit_module, [0, 1], [0, 1])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 0 has been initialized.
    phys_id0 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)
    # Check if qubit with virt ID 1 has been initialized.
    phys_id1 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=1)

    # Virtual qubit 0 should be in |1>
    qubit0 = processor.qdevice.get_local_qubit(phys_id0)
    assert has_state(qubit0, ketstates.s1)

    # Virtual qubit 1 should be in |0>
    qubit1 = processor.qdevice.get_local_qubit(phys_id1)
    assert has_state(qubit1, ketstates.s0)

    # New subroutine: apply X to q0 and Y to q1
    subrt = """
    x Q0
    y Q1
    """
    set_new_vanilla_subroutine(process, subrt, [0, 1], [0, 1])
    execute_process(processor, process)

    # q0 should be back in |0>
    qubit0 = processor.qdevice.get_local_qubit(phys_id0)
    assert has_state(qubit0, ketstates.s0)

    # q1 should be in |1>
    qubit1 = processor.qdevice.get_local_qubit(phys_id1)
    assert has_state(qubit1, ketstates.s1)

    # New subroutine: init q2, apply Y-rotation of PI/2 on q0
    # pi/2 = 8 / 2^4 * pi
    subrt = """
    set Q2 2
    init Q2
    rot_y Q0 8 4
    """
    set_new_vanilla_subroutine(process, subrt, [2], [2])
    execute_process(processor, process)

    # q0 should be in |+>
    qubit0 = processor.qdevice.get_local_qubit(phys_id0)
    assert has_state(qubit0, ketstates.h0)

    # q1 should still be in |1>
    qubit1 = processor.qdevice.get_local_qubit(phys_id1)
    assert has_state(qubit1, ketstates.s1)

    # Check if qubit with virt ID 2 has been initialized.
    phys_id2 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=2)

    # q2 should be in |0>
    qubit2 = processor.qdevice.get_local_qubit(phys_id2)
    assert has_state(qubit2, ketstates.s0)


def test_two_qubit_gates_generic():
    num_qubits = 3
    processor, unit_module = setup_components_generic(num_qubits)

    # Initialize q0 and q1. Apply CNOT between q0 and q1.
    subrt = """
    set Q0 0
    init Q0
    set Q1 1
    init Q1
    cnot Q0 Q1
    """

    process = create_process_with_vanilla_subrt(0, subrt, unit_module, [0, 1], [0, 1])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 0 has been initialized.
    phys_id0 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)
    # Check if qubit with virt ID 1 has been initialized.
    phys_id1 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=1)

    # Virtual qubit 0 should be in |0>
    q0 = processor.qdevice.get_local_qubit(phys_id0)
    assert has_state(q0, ketstates.s0)

    # Virtual qubit 1 should be in |0>
    q1 = processor.qdevice.get_local_qubit(phys_id1)
    assert has_state(q1, ketstates.s0)

    # New subroutine: apply H to q0 and again CNOT between q0 and q1
    subrt = """
    h Q0
    cnot Q0 Q1
    """
    set_new_vanilla_subroutine(process, subrt, [0, 1], [0, 1])
    execute_process(processor, process)

    # q0 and q1 should be maximally entangled
    [q0, q1] = processor.qdevice.get_local_qubits([phys_id0, phys_id1])
    # TODO: fix fidelity calculation with mixed states
    # assert has_max_mixed_state(q0)
    assert has_multi_state([q0, q1], ketstates.b00)

    # New subroutine: apply CNOT between q1 and q0
    subrt = """
    cnot Q1 Q0
    """
    set_new_vanilla_subroutine(process, subrt, [0, 1], [0, 1])
    execute_process(processor, process)

    # q0 should be |0>
    # q1 should be |+>
    [q0, q1] = processor.qdevice.get_local_qubits([phys_id0, phys_id1])
    assert has_state(q0, ketstates.s0)
    assert has_state(q1, ketstates.h0)


def test_multiple_processes_generic():
    num_qubits = 3
    processor, unit_module = setup_components_generic(num_qubits)

    # Process 0: initialize q0.
    subrt0 = """
    set Q0 0
    init Q0
    """

    # Process 1: initialize q0.
    subrt1 = """
    set Q0 0
    init Q0
    """

    process0 = create_process_with_vanilla_subrt(0, subrt0, unit_module, [0], [0])
    process1 = create_process_with_vanilla_subrt(1, subrt1, unit_module, [0], [0])
    processor._interface.memmgr.add_process(process0)
    processor._interface.memmgr.add_process(process1)
    execute_multiple_processes(processor, [process0, process1])

    # Check if qubit with virt ID 0 has been initialized for process 0
    proc0_phys_id0 = processor._interface.memmgr.phys_id_for(pid=0, virt_id=0)
    # Should be mapped to phys ID 0
    assert proc0_phys_id0 == 0

    # Check if qubit with virt ID 0 has been initialized for process 1
    proc1_phys_id0 = processor._interface.memmgr.phys_id_for(pid=1, virt_id=0)
    # Should be mapped to phys ID 1
    assert proc1_phys_id0 == 1

    # Process 0 virt qubit 0 should be in |0>
    proc0_q0 = processor.qdevice.get_local_qubit(proc0_phys_id0)
    assert has_state(proc0_q0, ketstates.s0)

    # Process 1 virt qubit 0 should be in |0>
    proc1_q0 = processor.qdevice.get_local_qubit(proc1_phys_id0)
    assert has_state(proc1_q0, ketstates.s0)

    # New subroutine for process 0: apply X to q0 and initialize q1
    subrt0 = """
    x Q0
    set Q1 1
    init Q1
    """
    set_new_vanilla_subroutine(process0, subrt0, [0, 1], [0, 1])

    # New subroutine for process 0: apply H to q0
    subrt1 = """
    h Q0
    """
    set_new_vanilla_subroutine(process1, subrt1, [0], [0])
    execute_multiple_processes(processor, [process0, process1])

    # Check if qubit with virt ID 1 has been initialized for process 0
    proc0_phys_id1 = processor._interface.memmgr.phys_id_for(pid=0, virt_id=1)
    # Should be mapped to phys ID 2
    assert proc0_phys_id1 == 2

    # Process 0 virt qubit 0 should be in |1>
    proc0_q0 = processor.qdevice.get_local_qubit(proc0_phys_id0)
    assert has_state(proc0_q0, ketstates.s1)

    # Process 0 virt qubit 1 should be in |0>
    proc0_q1 = processor.qdevice.get_local_qubit(proc0_phys_id1)
    assert has_state(proc0_q1, ketstates.s0)

    # Process 1 virt qubit 0 should be in |+>
    proc1_q0 = processor.qdevice.get_local_qubit(proc1_phys_id0)
    assert has_state(proc1_q0, ketstates.h0)

    # New subroutine for process 1: alloc q1
    subrt1 = """
    set Q1 1
    init Q1
    """
    set_new_vanilla_subroutine(process1, subrt1, [1], [1])
    # Should raise an AllocError since no physical qubits left.
    with pytest.raises(AllocError):
        execute_multiple_processes(processor, [process0, process1])

    # Free q0 for process 0
    processor.interface.memmgr.free(process0.pid, 0)
    # Try again same subroutine for process 1
    execute_process(processor, process1)

    # Check that qubit with virt ID 0 has been freed for process 0
    assert processor._interface.memmgr.phys_id_for(pid=0, virt_id=0) is None

    # Check that qubit with virt ID 1 for process 0 is still mapped to phys ID 2
    assert processor._interface.memmgr.phys_id_for(pid=0, virt_id=1) == 2

    # Check that qubit with virt ID 0 for process 1 is still mapped to phys ID 1
    assert processor._interface.memmgr.phys_id_for(pid=1, virt_id=0) == 1

    # Check that qubit with virt ID 1 for process 1 is now mapped to phys ID 0
    assert processor._interface.memmgr.phys_id_for(pid=1, virt_id=1) == 0

    # Check that physical qubit 0 has been reset to |0>
    # (because of initializing virt qubit 1 in process 1)
    phys_qubit_0 = processor.qdevice.get_local_qubit(0)
    assert has_state(phys_qubit_0, ketstates.s0)


def test_single_gates_nv_comm():
    num_qubits = 3

    # Set up a NV processor where
    # - comm qubit can only do ROT_X and ROT_Y
    # - mem qubit can do ROT_X, ROT_Y, ROT_Z
    processor, unit_module = setup_components_nv_star(num_qubits)

    subrt = """
    set Q0 0
    init Q0
    rot_x Q0 16 4
    """

    process = create_process_with_nv_subrt(0, subrt, unit_module, [0], [0])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 0 has been initialized.
    phys_id = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)

    # Qubit should be in |1>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s1)

    # New subroutine: apply rot_x on qubit 0 (comm).
    subrt = """
    rot_x Q0 16 4
    """
    set_new_nv_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should be back in |0>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)

    # New subroutine: apply rot_z on qubit 0 (comm), decomposed as rot_x and rot_y.
    subrt = """
    rot_x Q0 24 4
    rot_y Q0 16 4
    rot_x Q0 8 4
    """
    set_new_nv_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should still be in |0>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)

    # New subroutine: apply PI/2 Y-rotation.
    # pi/2 = 8 / 2^4 * pi
    subrt = """
    rot_y Q0 8 4
    """
    set_new_nv_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should be in |+>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.h0)

    # New subroutine: apply -PI/2 Z-rotation, decomposed as X-rotations and Y-rotation.
    # -pi/2 = 24 / 2^4 * pi
    subrt = """
    rot_x Q0 24 4
    rot_y Q0 24 4
    rot_x Q0 8 4
    """
    set_new_nv_subroutine(process, subrt, [0], [0])
    execute_process(processor, process)

    # Qubit should be in |-i>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.y1)


def test_single_gates_nv_mem():
    num_qubits = 3

    # Set up a NV processor where
    # - comm qubit can only do ROT_X and ROT_Y
    # - mem qubit can do ROT_X, ROT_Y, ROT_Z
    processor, unit_module = setup_components_nv_star(num_qubits)

    subrt = """
    set Q1 1
    init Q1
    rot_x Q1 16 4
    """

    process = create_process_with_nv_subrt(0, subrt, unit_module, [1], [1])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 1 has been initialized.
    phys_id = processor._interface.memmgr.phys_id_for(process.pid, virt_id=1)

    # Qubit should be in |1>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s1)

    # New subroutine: apply rot_x on qubit 1 (mem).
    subrt = """
    rot_x Q1 16 4
    """
    set_new_nv_subroutine(process, subrt, [1], [1])
    execute_process(processor, process)

    # Qubit should be back in |0>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)

    # New subroutine: apply rot_z on qubit 1 (mem). Should be allowed.
    subrt = """
    rot_z Q1 16 4
    """
    set_new_nv_subroutine(process, subrt, [1], [1])
    execute_process(processor, process)

    # Qubit should still be in |0>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.s0)

    # New subroutine: apply PI/2 Y-rotation.
    # pi/2 = 8 / 2^4 * pi
    subrt = """
    rot_y Q1 8 4
    """
    set_new_nv_subroutine(process, subrt, [1], [1])
    execute_process(processor, process)

    # Qubit should be in |+>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.h0)

    # New subroutine: apply -PI/2 Z-rotation.
    # -pi/2 = 24 / 2^4 * pi
    subrt = """
    rot_z Q1 24 4
    """
    set_new_nv_subroutine(process, subrt, [1], [1])
    execute_process(processor, process)

    # Qubit should be in |-i>
    qubit = processor.qdevice.get_local_qubit(phys_id)
    assert has_state(qubit, ketstates.y1)


def test_two_qubit_gates_nv():
    ns.sim_reset()
    num_qubits = 3
    processor, unit_module = setup_components_nv_star(num_qubits)

    # Initialize q0 and q1. Apply CROT_X (angle 0) between q0 and q1.
    subrt = """
    set Q0 0
    init Q0
    set Q1 1
    init Q1
    crot_x Q0 Q1 0 4
    """

    process = create_process_with_nv_subrt(0, subrt, unit_module, [0, 1], [0, 1])
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    # Check if qubit with virt ID 0 has been initialized.
    phys_id0 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=0)
    # Check if qubit with virt ID 1 has been initialized.
    phys_id1 = processor._interface.memmgr.phys_id_for(process.pid, virt_id=1)

    # Virtual qubit 0 should be in |0>
    q0 = processor.qdevice.get_local_qubit(phys_id0)
    assert has_state(q0, ketstates.s0)

    # Virtual qubit 1 should be in |0>
    q1 = processor.qdevice.get_local_qubit(phys_id1)
    assert has_state(q1, ketstates.s0)

    # New subroutine: apply CNOT between q0 and q1.
    # Should not be allowed because CNOT is not in NV flavour.
    subrt = """
    cnot Q0 Q1
    """
    # parse as vanilla just to check that NVProcessor cannot handle it
    # (parsing as NV will already give a parsing error)
    set_new_vanilla_subroutine(process, subrt, [0, 1], [0, 1])
    with pytest.raises(
        NotImplementedError
    ):  # NVProcessor does not implement _interpret_single_qubit_instr
        execute_process(processor, process)

    # New subroutine: apply H to q0 and CNOT between q0 and q1, decomposed using NV gates.
    subrt = """
    rot_y Q0 8 4
    rot_x Q0 16 4
    crot_x Q0 Q1 8 4
    rot_x Q0 24 4
    rot_y Q0 24 4
    rot_x Q0 8 4
    rot_x Q1 24 4
    """
    set_new_nv_subroutine(process, subrt, [0, 1], [0, 1])
    execute_process(processor, process)

    # q0 and q1 should be maximally entangled
    [q0, q1] = processor.qdevice.get_local_qubits([phys_id0, phys_id1])
    assert has_multi_state([q0, q1], ketstates.b00)

    # New subroutine: apply CNOT between q1 and q0, decomposed using NV gates.
    subrt = """
    rot_y Q0 8 4
    rot_x Q0 16 4
    rot_y Q1 8 4
    crot_x Q0 Q1 8 4
    rot_x Q0 24 4
    rot_y Q0 24 4
    rot_x Q0 8 4
    rot_y Q0 8 4
    rot_x Q0 16 4
    rot_x Q1 24 4
    rot_y Q1 24 4
    """
    set_new_nv_subroutine(process, subrt, [0, 1], [0, 1])
    execute_process(processor, process)

    # q0 should be |0>
    # q1 should be |+>
    [q0, q1] = processor.qdevice.get_local_qubits([phys_id0, phys_id1])
    assert has_state(q0, ketstates.s0)
    assert has_state(q1, ketstates.h0)


def test_multiple_processes_nv_alloc_error():
    num_qubits = 3
    processor, unit_module = setup_components_nv_star(num_qubits)

    # Process 0: initialize q0.
    subrt0 = """
    set Q0 0
    init Q0
    """

    # Process 1: initialize q0.
    subrt1 = """
    set Q0 0
    init Q0
    """

    process0 = create_process_with_nv_subrt(0, subrt0, unit_module, [0], [0])
    process1 = create_process_with_nv_subrt(1, subrt1, unit_module, [0], [0])
    processor._interface.memmgr.add_process(process0)
    processor._interface.memmgr.add_process(process1)

    # Should raise an AllocError: both processes allocate their virtual comm qubit
    # but there is only one physical comm qubit available.
    with pytest.raises(AllocError):
        execute_multiple_processes(processor, [process0, process1])


def test_initialize_all():
    num_qubits = 3
    processor, unit_module = setup_components_trapped_ion(num_qubits)

    subrt = """
    init_all
    """
    process = create_process_with_trapped_ion_subrt(
        0, subrt, unit_module, [0, 1, 2], [0, 1, 2]
    )
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)
    qubit2 = processor.qdevice.get_local_qubit(2)

    assert has_state(qubit0, ketstates.s0)
    assert has_state(qubit1, ketstates.s0)
    assert has_state(qubit2, ketstates.s0)


def test_rotate_all():
    num_qubits = 3
    processor, unit_module = setup_components_trapped_ion(num_qubits)

    subrt = """
        init_all
        rot_x_all 4 2
        """
    process = create_process_with_trapped_ion_subrt(
        0, subrt, unit_module, [0, 1, 2], [0, 1, 2]
    )
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)
    qubit2 = processor.qdevice.get_local_qubit(2)
    assert has_state(qubit0, ketstates.s1)
    assert has_state(qubit1, ketstates.s1)
    assert has_state(qubit2, ketstates.s1)

    subrt = """
    init_all
    rot_y_all 8 4
    """
    set_new_trapped_ion_subroutine(process, subrt, [0, 1, 2], [0, 1, 2])
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)
    qubit2 = processor.qdevice.get_local_qubit(2)
    assert has_state(qubit0, ketstates.h0)
    assert has_state(qubit1, ketstates.h0)
    assert has_state(qubit2, ketstates.h0)

    subrt = """
    rot_z_all 24 4
    """

    set_new_trapped_ion_subroutine(process, subrt, [0, 1, 2], [0, 1, 2])
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)
    qubit2 = processor.qdevice.get_local_qubit(2)

    assert has_state(qubit0, ketstates.y1)
    assert has_state(qubit1, ketstates.y1)
    assert has_state(qubit2, ketstates.y1)


def test_bichromatic_2_qubits():
    num_qubits = 2
    processor, unit_module = setup_components_trapped_ion(num_qubits)

    subrt = """
        set Q0 0
        set Q1 1
        init_all

        // rotate to |+> (note: not full hadamard!)
        rot_x_all 8 4
        rot_z Q0 8 4
        rot_x_all 24 4

        // cnot between q0 and q1
        rot_x_all 8 4
        rot_z Q0 8 4
        rot_x_all 24 4
        bichromatic 8 4
        rot_x_all 24 4
        rot_x_all 8 4
        rot_z Q0 24 4
        rot_x_all 24 4
        """
    process = create_process_with_trapped_ion_subrt(
        0, subrt, unit_module, [0, 1], [0, 1]
    )
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)

    assert qubit0 is not None
    assert qubit1 is not None

    # print(np.around(qubitapi.reduced_dm([qubit0, qubit1]), 2))
    assert has_multi_state([qubit0, qubit1], ketstates.b00)


def test_bichromatic_3_qubits():
    num_qubits = 3
    processor, unit_module = setup_components_trapped_ion(num_qubits)

    subrt = """
        init_all
        bichromatic 8 4
        """
    process = create_process_with_trapped_ion_subrt(
        0, subrt, unit_module, [0, 1, 2], [0, 1, 2]
    )
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)
    qubit2 = processor.qdevice.get_local_qubit(2)

    assert qubit0 is not None
    assert qubit1 is not None
    assert qubit2 is not None


def test_measure_all():
    num_qubits = 3
    processor, unit_module = setup_components_trapped_ion(num_qubits)

    subrt = """
            set Q0 0
            init_all
            meas_all Q0 @output
            """
    process = create_process_with_trapped_ion_subrt(
        0,
        subrt,
        unit_module,
        [0, 1, 2],
    )
    process.shared_mem.allocate_lr_out(3)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    qubit0 = processor.qdevice.get_local_qubit(0)
    qubit1 = processor.qdevice.get_local_qubit(1)
    qubit2 = processor.qdevice.get_local_qubit(2)

    assert process.shared_mem.read_lr_out(MemAddr(0), 3) == [0, 0, 0]
    assert qubit0 is not None
    assert qubit1 is not None
    assert qubit2 is not None


if __name__ == "__main__":
    test_init_qubit()
    test_init_qubit_with_latencies()
    test_init_not_allocated()
    test_single_gates_generic()
    test_single_gates_multiple_qubits_generic()
    test_two_qubit_gates_generic()
    test_multiple_processes_generic()
    test_single_gates_nv_comm()
    test_single_gates_nv_mem()
    test_two_qubit_gates_nv()
    test_multiple_processes_nv_alloc_error()
    test_initialize_all()
    test_rotate_all()
    test_bichromatic_2_qubits()
    test_bichromatic_3_qubits()



from __future__ import annotations

import logging
import random
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, Generator, List, Optional, Set, Tuple

import netsquid as ns
from netqasm.lang.operand import Template
from netsquid.components.cchannel import ClassicalChannel
from netsquid.components.component import Component, Port
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.lang import hostlang
from qoala.lang.ehi import (
    EhiNetworkInfo,
    EhiNetworkSchedule,
    EhiNetworkTimebin,
    EhiNodeInfo,
)
from qoala.lang.hostlang import BasicBlockType, ReceiveCMsgOp
from qoala.lang.request import VirtIdMappingType
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.message import Message
from qoala.runtime.program import (
    BatchInfo,
    BatchResult,
    ProgramBatch,
    ProgramInstance,
    ProgramResult,
)
from qoala.runtime.statistics import SchedulerStatistics
from qoala.runtime.task import (
    HostEventTask,
    LocalRoutineTask,
    MultiPairTask,
    ProcessorType,
    QoalaTask,
    SinglePairTask,
    TaskGraph,
    TaskGraphFromBlockBuilder,
    TaskInfo,
)
from qoala.sim.driver import CpuDriver, Driver, QpuDriver, SharedSchedulerMemory
from qoala.sim.eprsocket import EprSocket
from qoala.sim.events import EVENT_WAIT, SIGNAL_MEMORY_FREED, SIGNAL_TASK_COMPLETED
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.host import Host
from qoala.sim.host.hostinterface import HostInterface
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.netstack import Netstack
from qoala.sim.process import QoalaProcess
from qoala.sim.qnos import Qnos
from qoala.util.logging import LogManager


class NodeSchedulerComponent(Component):
    """
    NetSquid component representing for a node scheduler.
    It is used to send messages from the node scheduler to processor schedulers.

    :param name: Name of the component
    :param cpu_scheduler: CPU scheduler that node scheduler will send messages to.
    :param qpu_scheduler: QPU scheduler that node scheduler will send messages to.

    """

    def __init__(
        self,
        name,
        cpu_scheduler: ProcessorScheduler,
        qpu_scheduler: ProcessorScheduler,
        internal_sched_latency: float = 0.0,
    ):
        super().__init__(name=name)
        self.add_ports(["cpu_scheduler_out"])
        self.add_ports(["qpu_scheduler_out"])

        node_sched_to_cpu = ClassicalChannel(
            "node_scheduler_to_cpu_scheduler", delay=internal_sched_latency
        )
        self.cpu_scheduler_out_port.connect(node_sched_to_cpu.ports["send"])
        node_sched_to_cpu.ports["recv"].connect(cpu_scheduler.node_scheduler_in_port)
        node_sched_to_qpu = ClassicalChannel(
            "node_scheduler_to_qpu_scheduler", delay=internal_sched_latency
        )
        self.qpu_scheduler_out_port.connect(node_sched_to_qpu.ports["send"])
        node_sched_to_qpu.ports["recv"].connect(qpu_scheduler.node_scheduler_in_port)

    @property
    def cpu_scheduler_out_port(self) -> Port:
        """
        Port used to send messages to the CPU scheduler.
        """
        return self.ports["cpu_scheduler_out"]

    @property
    def qpu_scheduler_out_port(self) -> Port:
        """
        Port used to send messages to the QPU scheduler.
        """
        return self.ports["qpu_scheduler_out"]

    def send_cpu_scheduler_message(self, msg: Message) -> None:
        """
        Send a message to the CPU scheduler.
        :param msg: Message to send.
        :return: None
        """
        self.cpu_scheduler_out_port.tx_output(msg)

    def send_qpu_scheduler_message(self, msg: Message) -> None:
        """
        Send a message to the QPU scheduler.
        :param msg: Message to send.
        :return: None
        """
        self.qpu_scheduler_out_port.tx_output(msg)


class NodeScheduler(Protocol):
    def __init__(
        self,
        node_name: str,
        host: Host,
        qnos: Qnos,
        netstack: Netstack,
        memmgr: MemoryManager,
        local_ehi: EhiNodeInfo,
        network_ehi: EhiNetworkInfo,
        deterministic: bool = True,
        use_deadlines: bool = True,
        fcfs: bool = False,
        prio_epr: bool = False,
        is_predictable: bool = False,
    ) -> None:
        super().__init__(name=f"{node_name}_scheduler")

        self._node_name = node_name
        self.add_signal(SIGNAL_TASK_COMPLETED)

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({node_name})"
        )
        self._task_logger = LogManager.get_task_logger(f"{node_name}_NodeScheduler")

        self._host = host
        self._qnos = qnos
        self._netstack = netstack
        self._memmgr = memmgr
        self._local_ehi = local_ehi
        self._network_ehi = network_ehi

        self._prog_instance_counter: int = 0
        self._batch_counter: int = 0
        self._batches: Dict[int, ProgramBatch] = {}  # batch ID -> batch
        self._prog_results: Dict[int, ProgramResult] = {}  # program ID -> result
        self._batch_results: Dict[int, BatchResult] = {}  # batch ID -> result

        self._task_counter = 0
        self._task_graph: Optional[TaskGraph] = None
        self._prog_start_timestamps: Dict[int, float] = {}  # program ID -> start time
        self._prog_end_timestamps: Dict[int, float] = {}  # program ID -> end time

        self._current_block_index: Dict[int, int] = {}  # program ID -> block index
        self._task_from_block_builder = TaskGraphFromBlockBuilder()
        self._prog_instance_dependency: Dict[
            int, int
        ] = {}  # program ID -> dependent program ID

        self._const_batch: Optional[ProgramBatch] = None

        self._last_cpu_task_pid = -1
        self._last_qpu_task_pid = -1

        scheduler_memory = SharedSchedulerMemory()
        netschedule = network_ehi.network_schedule

        self._is_predictable = is_predictable

        # TODO: refactor
        node_id = self.host._comp.node_id

        cpudriver = CpuDriver(node_name, scheduler_memory, host.processor, memmgr)
        cpu_sched_typ = CpuFcfsScheduler if fcfs else CpuEdfScheduler
        self._cpu_scheduler: CpuScheduler = cpu_sched_typ(  # type: ignore
            f"{node_name}_cpu",
            node_id,
            cpudriver,
            memmgr,
            host.interface,
            deterministic,
            use_deadlines,
        )

        qpudriver = QpuDriver(
            node_name,
            scheduler_memory,
            qnos.processor,
            netstack.processor,
            memmgr,
        )
        self._qpu_scheduler = QpuScheduler(
            f"{node_name}_qpu",
            node_id,
            qpudriver,
            memmgr,
            netschedule,
            deterministic,
            use_deadlines,
            prio_epr,
        )

        self._comp = NodeSchedulerComponent(
            f"{node_name}_scheduler",
            self._cpu_scheduler,
            self._qpu_scheduler,
            internal_sched_latency=local_ehi.latencies.internal_sched_latency,
        )

        self._cpu_scheduler.set_other_scheduler(self._qpu_scheduler)
        self._qpu_scheduler.set_other_scheduler(self._cpu_scheduler)

    @property
    def host(self) -> Host:
        return self._host

    @property
    def qnos(self) -> Qnos:
        return self._qnos

    @property
    def netstack(self) -> Netstack:
        return self._netstack

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    @property
    def cpu_scheduler(self) -> ProcessorScheduler:
        return self._cpu_scheduler

    @property
    def qpu_scheduler(self) -> ProcessorScheduler:
        return self._qpu_scheduler

    def submit_batch(self, batch_info: BatchInfo) -> ProgramBatch:
        prog_instances: List[ProgramInstance] = []

        for i in range(batch_info.num_iterations):
            pid = self._prog_instance_counter

            instance = ProgramInstance(
                pid=pid,
                program=batch_info.program,
                inputs=batch_info.inputs[i],
                unit_module=batch_info.unit_module,
            )
            self._prog_instance_counter += 1
            prog_instances.append(instance)

        batch = ProgramBatch(
            batch_id=self._batch_counter, info=batch_info, instances=prog_instances
        )
        self._batches[batch.batch_id] = batch
        self._batch_counter += 1
        return batch

    def submit_const_batch(self, batch_info: BatchInfo) -> ProgramBatch:
        prog_instances: List[ProgramInstance] = []

        for i in range(batch_info.num_iterations):
            pid = self._prog_instance_counter

            instance = ProgramInstance(
                pid=pid,
                program=batch_info.program,
                inputs=batch_info.inputs[i],
                unit_module=batch_info.unit_module,
            )
            self._prog_instance_counter += 1
            prog_instances.append(instance)

        batch = ProgramBatch(
            batch_id=self._batch_counter, info=batch_info, instances=prog_instances
        )
        self._const_batch = batch
        return batch

    def get_batches(self) -> Dict[int, ProgramBatch]:
        return self._batches

    def create_process(
        self, prog_instance: ProgramInstance, remote_pid: Optional[int] = None
    ) -> QoalaProcess:
        prog_memory = ProgramMemory(prog_instance.pid)
        meta = prog_instance.program.meta

        csockets: Dict[int, ClassicalSocket] = {}
        for i, remote_name in meta.csockets.items():
            assert remote_pid is not None
            # TODO: check for already existing classical sockets
            csockets[i] = self.host.create_csocket(
                remote_name, prog_instance.pid, remote_pid
            )

        epr_sockets: Dict[int, EprSocket] = {}
        for i, remote_name in meta.epr_sockets.items():
            assert remote_pid is not None
            remote_id = self._network_ehi.get_node_id(remote_name)
            # TODO: check for already existing epr sockets
            # TODO: fidelity
            epr_sockets[i] = EprSocket(i, remote_id, prog_instance.pid, remote_pid, 1.0)

        result = ProgramResult(values={})

        return QoalaProcess(
            prog_instance=prog_instance,
            prog_memory=prog_memory,
            csockets=csockets,
            epr_sockets=epr_sockets,
            result=result,
        )

    def create_processes_for_batches(
        self,
        remote_pids: Optional[Dict[int, List[int]]] = None,  # batch ID -> PID list
        linear: bool = False,
    ) -> None:
        prev_prog_instance_id = -1
        for batch_id, batch in self._batches.items():
            for i, prog_instance in enumerate(batch.instances):
                if remote_pids is not None and batch_id in remote_pids:
                    remote_pid = remote_pids[batch_id][i]
                else:
                    remote_pid = None
                process = self.create_process(prog_instance, remote_pid)

                self.memmgr.add_process(process)
                self.initialize_process(process)
                self._current_block_index[prog_instance.pid] = 0
                if linear:
                    self._prog_instance_dependency[
                        prog_instance.pid
                    ] = prev_prog_instance_id
                    prev_prog_instance_id = prog_instance.pid
                else:
                    self._prog_instance_dependency[prog_instance.pid] = -1

        if self._const_batch is not None:
            for i, prog_instance in enumerate(self._const_batch.instances):
                process = self.create_process(prog_instance)
                self.memmgr.add_process(process)
                self.initialize_process(process)

    def collect_timestamps(self, batch_id: int) -> List[Optional[Tuple[float, float]]]:
        batch = self._batches[batch_id]
        timestamps: List[Optional[Tuple[float, float]]] = []
        for prog_instance in batch.instances:
            process = self.memmgr.get_process(prog_instance.pid)
            cpu_start_end = self.cpu_scheduler.get_timestamps(process.pid)
            if cpu_start_end is None:
                timestamps.append(None)
                continue

            cpu_start, cpu_end = cpu_start_end
            qpu_start_end = self.qpu_scheduler.get_timestamps(process.pid)
            # QPU timestamps could be None (if program did not have any quantum tasks)
            if qpu_start_end is not None:
                qpu_start, qpu_end = qpu_start_end
                start = min(cpu_start, qpu_start)
                end = max(cpu_end, qpu_end)
                timestamps.append((start, end))
            else:
                timestamps.append((cpu_start, cpu_end))
        return timestamps

    def collect_batch_results(self) -> None:
        for batch_id, batch in self._batches.items():
            results: List[ProgramResult] = []
            for prog_instance in batch.instances:
                process = self.memmgr.get_process(prog_instance.pid)
                results.append(process.result)
            timestamps = self.collect_timestamps(batch_id)
            self._batch_results[batch_id] = BatchResult(batch_id, results, timestamps)

    def get_batch_results(self) -> Dict[int, BatchResult]:
        self.collect_batch_results()
        return self._batch_results

    def get_all_non_const_pids(self) -> List[int]:
        pids = self.memmgr.get_all_program_ids()
        if self._const_batch is None:
            return pids
        else:
            const_pids = [inst.pid for inst in self._const_batch.instances]
            return [pid for pid in pids if pid not in const_pids]

    def is_from_const_batch(self, pid: int) -> bool:
        if self._const_batch is None:
            return False
        # else:
        const_pids = [inst.pid for inst in self._const_batch.instances]
        return pid in const_pids

    def initialize_process(self, process: QoalaProcess) -> None:
        # Write program inputs to host memory.
        self.host.processor.initialize(process)

        inputs = process.prog_instance.inputs
        for req in process.prog_instance.program.request_routines.values():
            req.instantiate(inputs.values)

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def start(self) -> None:
        # Processor schedulers start first to ensure that they will start running tasks after they receive the first
        # message from the node scheduler.
        self._cpu_scheduler.start()
        self._qpu_scheduler.start()
        if not self._is_predictable:
            super().start()
            self.schedule_all()

    def stop(self) -> None:
        self._qpu_scheduler.stop()
        self._cpu_scheduler.stop()
        super().stop()

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            if (
                self._last_cpu_task_pid != -1
                and (
                    self.is_from_const_batch(self._last_cpu_task_pid)
                    or self.is_program_instance_finished(self._last_cpu_task_pid)
                )
            ) or (
                self._last_qpu_task_pid != -1
                and (
                    self.is_from_const_batch(self._last_qpu_task_pid)
                    or self.is_program_instance_finished(self._last_qpu_task_pid)
                )
            ):
                self.schedule_all()
            ev_expr = self.await_signal(self._cpu_scheduler, SIGNAL_TASK_COMPLETED)
            ev_expr = ev_expr | self.await_signal(
                self._qpu_scheduler, SIGNAL_TASK_COMPLETED
            )
            yield ev_expr

            now = ns.sim_time()

            # Gets the pid of the last finished task at the current time,
            # if there is no task that is finished at the current time, it returns -1
            self._last_cpu_task_pid = self.cpu_scheduler.get_last_finished_task_pid_at(
                now
            )
            # If there is a task that is finished at the current time, assign the next
            if self._last_cpu_task_pid != -1 and not self.is_from_const_batch(
                self._last_cpu_task_pid
            ):
                self.schedule_next_for(self._last_cpu_task_pid)

            self._last_qpu_task_pid = self.qpu_scheduler.get_last_finished_task_pid_at(
                now
            )
            # If there is a task that is finished at the current time, assign the next, but we do not need to assign
            # again if it is the same pid as the one that came from CPU
            if (
                self._last_qpu_task_pid != -1
                and self._last_qpu_task_pid != self._last_cpu_task_pid
                and not self.is_from_const_batch(self._last_qpu_task_pid)
            ):
                self.schedule_next_for(self._last_qpu_task_pid)

    def schedule_next_for(self, pid: int) -> None:
        """
        Schedule the tasks of the next block for program instance with given pid
        by assigning respective tasks to CPU and QPU schedulers and send a message
        to schedulers for informing them about the newly assigned tasks.

        :param pid: program instance id
        :return: None
        """
        new_cpu_tasks, new_qpu_tasks = self.find_next_tasks_for(pid)

        # If there are new tasks, send a message to schedulers
        # Note that find_next_tasks_for() returns None if there are no new tasks for that processor
        if new_cpu_tasks:
            self._cpu_scheduler.add_tasks(new_cpu_tasks)
            self._comp.send_cpu_scheduler_message(Message(-1, -1, "New Task"))
        if new_qpu_tasks:
            self._qpu_scheduler.add_tasks(new_qpu_tasks)
            self._task_logger.debug(f"sending 'New Task' msg to QPU scheduler")
            self._comp.send_qpu_scheduler_message(Message(-1, -1, "New Task"))

    def schedule_all(self) -> None:
        """
        Schedules the tasks of the next block for each available program instance in the memory manager
        by assigning respective tasks to CPU and QPU schedulers and sends a message
        to schedulers for informing them about the newly assigned tasks.

        This method is responsible for scheduling tasks for each available program instance in the memory manager.
        A program instance is considered available if it meets two conditions:
        1. It is not finished.
        2. It does not have any dependencies on an unfinished program instance
        (it can have such dependencies if the batch of program instances are submitted to run linearly).

        :return: None
        """
        all_new_cpu_tasks: Dict[int, TaskInfo] = {}
        all_new_qpu_tasks: Dict[int, TaskInfo] = {}

        # for pid in self.memmgr.get_all_program_ids():
        for pid in self.get_all_non_const_pids():
            # If there is a dependency, check if it is finished
            dependency_pid = self._prog_instance_dependency[pid]
            if dependency_pid != -1:
                dep_cur_index = self._current_block_index[dependency_pid]
                dep_block_length = len(
                    self.memmgr.get_process(dependency_pid).prog_instance.program.blocks
                )
                if dep_cur_index < dep_block_length:
                    continue

            # Note that find_next_tasks_for() returns None if there are no new tasks for that processor
            new_cpu_tasks, new_qpu_tasks = self.find_next_tasks_for(pid)
            if new_cpu_tasks:
                all_new_cpu_tasks.update(new_cpu_tasks)
            if new_qpu_tasks:
                all_new_qpu_tasks.update(new_qpu_tasks)

        # If there are new tasks, send a message to schedulers
        if len(all_new_cpu_tasks) > 0:
            self._cpu_scheduler.add_tasks(all_new_cpu_tasks)
            self._comp.send_cpu_scheduler_message(Message(-1, -1, "New Task"))
        if len(all_new_qpu_tasks) > 0:
            self._qpu_scheduler.add_tasks(all_new_qpu_tasks)
            self._task_logger.debug(f"sending 'New Task' msg to QPU scheduler")
            self._comp.send_qpu_scheduler_message(Message(-1, -1, "New Task"))

    def find_next_tasks_for(
        self, pid: int
    ) -> Tuple[Optional[Dict[int, TaskInfo]], Optional[Dict[int, TaskInfo]]]:
        """
        Finds the tasks of the next block for program instance with given pid,
        and returns them as CPU tasks and QPU tasks separately.

        :param pid: The program instance ID for which to find the next tasks.
        :return: A 2-tuple containing the new CPU tasks and new QPU tasks. If no tasks are
             found for the given PID, both elements of the tuple will be set to None.
        """
        new_cpu_tasks: Dict[int, TaskInfo] = {}
        new_qpu_tasks: Dict[int, TaskInfo] = {}

        if (
            pid in self.host.interface.program_instance_jumps
            and self.host.interface.program_instance_jumps[pid] != -1
        ):
            self._current_block_index[pid] = self.host.interface.program_instance_jumps[
                pid
            ]
            self.host.interface.program_instance_jumps[pid] = -1

        current_block_index = self._current_block_index[pid]
        prog_instance = self.memmgr.get_process(pid).prog_instance
        blocks = prog_instance.program.blocks

        is_program_finished = current_block_index >= len(blocks)
        # If program is finished or CPU scheduler has a task for this pid, do not schedule
        # Note that for all block types, there will be tasks for CPU scheduler
        if is_program_finished or self.cpu_scheduler.task_exists_for_pid(pid):
            return None, None

        block = prog_instance.program.blocks[current_block_index]

        # If it is CL or CC it will only have tasks in CPU but others will have tasks in both
        if block.typ in {
            BasicBlockType.CL,
            BasicBlockType.CC,
        }:
            graph = self._task_from_block_builder.build(
                prog_instance, current_block_index, self._network_ehi
            )
            new_cpu_tasks.update(graph.get_tasks())

            self._current_block_index[pid] += 1
            # If scheduler does not have any task send a message to wake it up
            self._comp.send_cpu_scheduler_message(Message(-1, -1, "New Task"))
            return new_cpu_tasks, None
        elif not self.qpu_scheduler.task_exists_for_pid(pid):
            # Note that we know that cpu does not have any tasks for this pid
            graph = self._task_from_block_builder.build(
                prog_instance, current_block_index, self._network_ehi
            )
            cpu_graph = graph.partial_graph(ProcessorType.CPU).get_tasks()
            qpu_graph = graph.partial_graph(ProcessorType.QPU).get_tasks()
            self._task_logger.debug(f"adding CPU tasks {cpu_graph}")
            self._task_logger.debug(f"adding QPU tasks {qpu_graph}")
            new_cpu_tasks.update(cpu_graph)
            new_qpu_tasks.update(qpu_graph)

            self._current_block_index[pid] += 1

        return new_cpu_tasks, new_qpu_tasks

    def upload_task_graph(self, graph: TaskGraph) -> None:
        """
        Assigns tasks in the given task graph to the CPU and QPU schedulers.

        :param graph: The task graph to upload.
        :return: None
        """
        self._task_graph = graph
        cpu_graph = graph.partial_graph(ProcessorType.CPU)
        qpu_graph = graph.partial_graph(ProcessorType.QPU)
        self._cpu_scheduler.upload_task_graph(cpu_graph)
        self._qpu_scheduler.upload_task_graph(qpu_graph)

    def is_program_instance_finished(self, pid: int) -> bool:
        return self._current_block_index[pid] >= len(
            self.memmgr.get_process(pid).prog_instance.program.blocks
        )

    def submit_program_instance(
        self, prog_instance: ProgramInstance, remote_pid: Optional[int] = None
    ) -> None:
        process = self.create_process(prog_instance, remote_pid)
        self.memmgr.add_process(process)
        self.initialize_process(process)
        self._current_block_index[prog_instance.pid] = 0
        self._prog_instance_dependency[prog_instance.pid] = -1

    def get_statistics(self) -> SchedulerStatistics:
        return SchedulerStatistics(
            cpu_tasks_executed=self.cpu_scheduler.get_tasks_executed(),
            qpu_tasks_executed=self.qpu_scheduler.get_tasks_executed(),
            cpu_task_starts=self.cpu_scheduler.get_task_starts(),
            qpu_task_starts=self.qpu_scheduler.get_task_starts(),
            cpu_task_ends=self.cpu_scheduler.get_task_ends(),
            qpu_task_ends=self.qpu_scheduler.get_task_ends(),
        )


class ProcessorSchedulerComponent(Component):
    """
    NetSquid component representing for the ProcessorScheduler.
    It is used to receive messages from the node scheduler.

    :param name: Name of the component
    """

    def __init__(self, name):
        super().__init__(name=name)
        self.add_ports(["node_scheduler_in"])

    @property
    def node_scheduler_in_port(self) -> Port:
        """
        Port that the node scheduler uses to send messages to this component.
        """
        return self.ports["node_scheduler_in"]


class ProcessorScheduler(Protocol):
    def __init__(
        self,
        name: str,
        node_id: int,
        driver: Driver,
        memmgr: MemoryManager,
        deterministic: bool = True,
        use_deadlines: bool = True,
    ) -> None:
        super().__init__(name=name)
        self.add_signal(SIGNAL_TASK_COMPLETED)

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}_{driver.__class__.__name__}({name})"
        )
        self._node_id = node_id
        self._task_logger = LogManager.get_task_logger(name)
        self._driver = driver
        self._other_scheduler: Optional[ProcessorScheduler] = None
        self._memmgr = memmgr
        self._deterministic = deterministic
        self._use_deadlines = use_deadlines

        self._task_graph: TaskGraph = TaskGraph()
        self._finished_tasks: List[int] = []

        self._prog_start_timestamps: Dict[int, float] = {}  # program ID -> start time
        self._prog_end_timestamps: Dict[int, float] = {}  # program ID -> end time

        self._tasks_executed: Dict[int, QoalaTask] = {}
        self._task_starts: Dict[int, float] = {}
        self._task_ends: Dict[int, float] = {}
        self.last_finished_task_pid: Tuple[int, int] = (-1, -1)  # (pid, end_time)

        self._comp = ProcessorSchedulerComponent(name + "_comp")

        self._status: SchedulerStatus = SchedulerStatus(status=set(), params={})

    @property
    def node_scheduler_in_port(self) -> Port:
        return self._comp.node_scheduler_in_port

    @property
    def driver(self) -> Driver:
        return self._driver

    @property
    def status(self) -> SchedulerStatus:
        return self._status

    def update_external_predcessors(self) -> None:
        if self._other_scheduler is None:
            return
        assert self._task_graph is not None

        tg = self._task_graph

        for r in tg.get_roots(ignore_external=True):
            ext_preds = tg.get_tinfo(r).ext_predecessors
            new_ext_preds = {
                ext for ext in ext_preds if not self._other_scheduler.has_finished(ext)
            }
            tg.get_tinfo(r).ext_predecessors = new_ext_preds

    def upload_task_graph(self, graph: TaskGraph) -> None:
        """
        Sets the given task graph as the current task graph.

        :param graph: The task graph to upload.
        :return: None
        """
        self._task_graph = graph

    # Gets the pid of the last finished task at the current time,
    # if there is no task that is finished at the current time, it returns -1
    def get_last_finished_task_pid_at(self, time: float) -> int:
        """
        Finds the pid of the last finished task at the given time and returns it,
        if there is no task that is finished at the current time, it returns -1

        :param time: The time to check for finished tasks.
        :return: The pid of the last finished task at the given time if such task exists, -1 otherwise.
        """
        if self.last_finished_task_pid[1] == time:
            return self.last_finished_task_pid[0]
        else:
            return -1

    def task_exists_for_pid(self, pid: int) -> bool:
        """
        Checks the current task graph for the existence of a task with the given pid. Returns True if such task exists,
        False otherwise.

        :param pid: The pid to check for.
        :return: True if a task with the given pid exists, False otherwise.
        """
        return self._task_graph.task_exists_for_pid(pid)

    def add_tasks(self, tasks: Dict[int, TaskInfo]) -> None:
        """
        Adds the given tasks to the current task graph.

        :param tasks: The tasks to add.
        :return: None
        """
        self._task_graph.get_tasks().update(tasks)

    def has_finished(self, task_id: int) -> bool:
        return task_id in self._finished_tasks

    def set_other_scheduler(self, other: ProcessorScheduler) -> None:
        self._other_scheduler = other

    def record_start_timestamp(self, pid: int, time: float) -> None:
        # Only write start time for first encountered task.
        if pid not in self._prog_start_timestamps:
            self._prog_start_timestamps[pid] = time

    def record_end_timestamp(self, pid: int, time: float) -> None:
        # Overwrite end time for every task. Automatically the last timestamp remains.
        self._prog_end_timestamps[pid] = time

    def get_timestamps(self, pid: int) -> Optional[Tuple[float, float]]:
        if pid not in self._prog_start_timestamps:
            assert pid not in self._prog_end_timestamps
            return None
        assert pid in self._prog_end_timestamps
        return self._prog_start_timestamps[pid], self._prog_end_timestamps[pid]

    def get_tasks_executed(self) -> Dict[int, QoalaTask]:
        return self._tasks_executed

    def get_task_starts(self) -> Dict[int, float]:
        return self._task_starts

    def get_task_ends(self) -> Dict[int, float]:
        return self._task_ends

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def handle_task(self, task_id: int) -> Generator[EventExpression, None, None]:
        assert self._task_graph is not None
        tinfo = self._task_graph.get_tinfo(task_id)
        task = tinfo.task

        self._logger.debug(f"{ns.sim_time()}: {self.name}: checking next task {task}")

        before = ns.sim_time()

        start_time = self._task_graph.get_tinfo(task.task_id).start_time
        is_busy_task = start_time is not None
        self._logger.info(f"executing task {task}")
        if is_busy_task:
            self._task_logger.info(f"BUSY start  {task} (start time: {start_time})")
        else:
            self._task_logger.info(f"start  {task}")
        if self.name == "bob_qpu":
            self._task_logger.warning(f"start  {task}")
        elif self.name == "bob_cpu":
            if isinstance(task, HostEventTask):
                self._task_logger.warning(f"start  {task}")
        self._task_starts[task.task_id] = before
        self.record_start_timestamp(task.pid, before)

        # Execute the task
        success = yield from self._driver.handle_task(task)
        if success:
            after = ns.sim_time()

            self.record_end_timestamp(task.pid, after)
            self.last_finished_task_pid = (task.pid, after)
            duration = after - before
            self._task_graph.decrease_deadlines(duration)
            self._task_graph.remove_task(task_id)

            self._finished_tasks.append(task.task_id)
            self.send_signal(SIGNAL_TASK_COMPLETED)
            self._logger.info(f"finished task {task}")
            if is_busy_task:
                self._task_logger.info(f"BUSY finish {task}")
            else:
                self._task_logger.info(f"finish {task}")
            if self.name == "bob_qpu":
                self._task_logger.warning(f"finish {task}")

            self._tasks_executed[task.task_id] = task
            self._task_ends[task.task_id] = after
        else:
            self._task_logger.info("task failed")


class Status(Enum):
    GRAPH_EMPTY = auto()
    EPR_GEN = auto()
    NEXT_TASK = auto()
    WAITING_OTHER_CORE = auto()
    WAITING_MSG = auto()
    WAITING_START_TIME = auto()
    WAITING_RESOURCES = auto()
    WAITING_TIME_BIN = auto()


@dataclass
class SchedulerStatus:
    status: Set[Status]
    params: Dict[str, Any]


class CpuScheduler(ProcessorScheduler):
    def __init__(
        self,
        name: str,
        node_id: int,
        driver: CpuDriver,
        memmgr: MemoryManager,
        host_interface: HostInterface,
        deterministic: bool = True,
        use_deadlines: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            node_id=node_id,
            driver=driver,
            memmgr=memmgr,
            deterministic=deterministic,
            use_deadlines=use_deadlines,
        )
        self._host_interface = host_interface

    def is_message_available(self, tid: int) -> bool:
        assert self._task_graph is not None
        task = self._task_graph.get_tinfo(tid).task
        assert isinstance(task, HostEventTask)
        process = self._memmgr.get_process(task.pid)
        block = process.program.get_block(task.block_name)
        instr = block.instructions[0]
        assert isinstance(instr, ReceiveCMsgOp)
        assert isinstance(instr.arguments[0], hostlang.IqoalaSingleton)
        csck_id = process.host_mem.read(instr.arguments[0].name)
        csck = process.csockets[csck_id]
        remote_name = csck.remote_name
        remote_pid = csck.remote_pid
        self._task_logger.debug(f"checking if msg from {remote_name} is available")
        messages = self._host_interface.get_available_messages(remote_name)
        if (remote_pid, task.pid) in messages:
            self._task_logger.debug(f"task {tid} NOT blocked on message")
            return True
        else:
            self._task_logger.debug(f"task {tid} blocked on message")
            return False

    @abstractmethod
    def choose_next_task(self, ready_tasks: List[int]) -> None:
        raise NotImplementedError

    def update_status(self) -> None:
        tg = self._task_graph

        if tg is None or len(tg.get_tasks()) == 0:
            self._status = SchedulerStatus(status={Status.GRAPH_EMPTY}, params={})
            return

        # All tasks that have no predecessors, internal nor external.
        no_predecessors = tg.get_roots()

        # All tasks that have only external predecessors.
        blocked_on_other_core = tg.get_tasks_blocked_only_on_external()

        # All "receive message" tasks without predecessors (internal nor external).
        event_no_predecessors = [
            tid for tid in no_predecessors if tg.get_tinfo(tid).task.is_event_task()
        ]

        event_blocked_on_message = [
            tid for tid in event_no_predecessors if not self.is_message_available(tid)
        ]
        self._task_logger.info(f"event_blocked_on_message: {event_blocked_on_message}")

        now = ns.sim_time()
        with_future_start: Dict[int, float] = {
            tid: tg.get_tinfo(tid).start_time  # type: ignore
            for tid in no_predecessors
            if tg.get_tinfo(tid).start_time is not None
            and tg.get_tinfo(tid).start_time > now
        }
        wait_for_start: Optional[Tuple[int, float]] = None  # (task ID, start time)
        if len(with_future_start) > 0:
            sorted_by_start = sorted(
                with_future_start.items(), key=lambda item: item[1]
            )
            wait_for_start = sorted_by_start[0]
        self._task_logger.info(f"wait_for_start: {wait_for_start}")

        ready = [
            tid
            for tid in no_predecessors
            if tid not in event_blocked_on_message and tid not in with_future_start
        ]
        ready_task_dict = {tid: str(tg.get_tinfo(tid).task) for tid in ready}
        self._task_logger.debug(f"ready tasks: {ready}\n{ready_task_dict}")

        if len(ready) > 0:
            # self._task_logger.warning(f"ready tasks: {ready}")
            # From the readily executable tasks, choose which one to execute
            self.choose_next_task(ready)
        else:
            if len(blocked_on_other_core) > 0:
                self._logger.debug("Waiting other core")
                self._task_logger.debug("Waiting other core")
                self._status.status.add(Status.WAITING_OTHER_CORE)
            if len(event_blocked_on_message) > 0:
                self._logger.debug("Waiting message")
                self._task_logger.debug("Waiting message")
                self._status.status.add(Status.WAITING_MSG)
            if wait_for_start is not None:
                _, start = wait_for_start
                self._logger.debug("Waiting Start Time")
                self._task_logger.debug("Waiting Start Time")
                self._status.status.add(Status.WAITING_START_TIME)
                self._status.params["start_time"] = start

            if len(self.status.status) == 0:
                raise RuntimeError

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            self._task_logger.debug("updating status...")
            self._status = SchedulerStatus(status=set(), params={})
            self.update_external_predcessors()
            self.update_status()
            self._task_logger.debug(f"status: {self.status.status}")
            if Status.NEXT_TASK in self.status.status:
                task_id = self.status.params["task_id"]
                yield from self.handle_task(task_id)
            else:
                ev_expr = self.await_port_input(self.node_scheduler_in_port)
                if Status.WAITING_OTHER_CORE in self.status.status:
                    ev_expr = ev_expr | self.await_signal(
                        sender=self._other_scheduler,
                        signal_label=SIGNAL_TASK_COMPLETED,
                    )
                if Status.WAITING_START_TIME in self.status.status:
                    start_time = self.status.params["start_time"]
                    now = ns.sim_time()
                    delta = start_time - now
                    self._schedule_after(delta, EVENT_WAIT)
                    ev_start_time = EventExpression(source=self, event_type=EVENT_WAIT)
                    ev_expr = ev_expr | ev_start_time

                if Status.WAITING_MSG in self.status.status:
                    ev_msg_arrived = self._host_interface.get_evexpr_for_any_msg()

                    ev_expr = ev_msg_arrived | ev_expr
                    yield ev_expr
                    if len(ev_expr.first_term.triggered_events) > 0:
                        # It was "ev_msg_arrived" that triggered.
                        # Need to process this event (flushing potential other messages)
                        yield from self._host_interface.handle_msg_evexpr(
                            ev_expr.first_term
                        )
                else:
                    yield ev_expr


class CpuEdfScheduler(CpuScheduler):
    def __init__(
        self,
        name: str,
        node_id: int,
        driver: CpuDriver,
        memmgr: MemoryManager,
        host_interface: HostInterface,
        deterministic: bool = True,
        use_deadlines: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            node_id=node_id,
            driver=driver,
            memmgr=memmgr,
            host_interface=host_interface,
            deterministic=deterministic,
            use_deadlines=use_deadlines,
        )

    def choose_next_task(self, ready_tasks: List[int]) -> None:
        tg = self._task_graph

        with_deadline = [
            t
            for t in ready_tasks
            if tg.get_tinfo(t).deadline is not None
            or len(tg.get_tinfo(t).rel_deadlines) > 0
            or len(tg.get_tinfo(t).ext_rel_deadlines) > 0
        ]
        if not self._use_deadlines:
            with_deadline = []

        self._task_logger.debug(f"ready tasks with deadline: {with_deadline}")

        to_return: int

        if len(with_deadline) > 0:
            # Sort them by deadline and return the one with the earliest deadline
            deadlines = {t: tg.get_tinfo(t).deadline for t in with_deadline}
            sorted_by_deadline = sorted(deadlines.items(), key=lambda item: item[1])  # type: ignore
            if self.name in ["bob_cpu", "bob_qpu"]:
                self._task_logger.info(f"tasks with deadlines: {sorted_by_deadline}")
            to_return = sorted_by_deadline[0][0]
            self._logger.debug(f"Return task {to_return}")
            self._task_logger.debug(f"Return task {to_return}")
        else:
            # No deadlines
            if self._deterministic:
                to_return = ready_tasks[0]
            else:
                index = random.randint(0, len(ready_tasks) - 1)
                to_return = ready_tasks[index]
            self._logger.debug(f"Return task {to_return}")
            self._task_logger.debug(f"Return task {to_return}")
        self._status = SchedulerStatus(
            status={Status.NEXT_TASK}, params={"task_id": to_return}
        )


class CpuFcfsScheduler(CpuScheduler):
    def __init__(
        self,
        name: str,
        node_id: int,
        driver: CpuDriver,
        memmgr: MemoryManager,
        host_interface: HostInterface,
        deterministic: bool = True,
        use_deadlines: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            node_id=node_id,
            driver=driver,
            memmgr=memmgr,
            host_interface=host_interface,
            deterministic=deterministic,
            use_deadlines=use_deadlines,
        )

        self._task_queue: List[int] = []  # list of task IDs

    def choose_next_task(self, ready_tasks: List[int]) -> None:
        for tid in ready_tasks:
            if tid not in self._task_queue:
                self._task_queue.append(tid)

        self._task_logger.debug(f"task queue: {self._task_queue}")
        next_task = self._task_queue.pop(0)
        self._task_logger.debug(f"popping: {next_task}")

        self._status = SchedulerStatus(
            status={Status.NEXT_TASK}, params={"task_id": next_task}
        )


class QpuScheduler(ProcessorScheduler):
    def __init__(
        self,
        name: str,
        node_id: int,
        driver: QpuDriver,
        memmgr: MemoryManager,
        network_schedule: Optional[EhiNetworkSchedule] = None,
        deterministic: bool = True,
        use_deadlines: bool = True,
        prio_epr: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            node_id=node_id,
            driver=driver,
            memmgr=memmgr,
            deterministic=deterministic,
            use_deadlines=use_deadlines,
        )
        self._network_schedule = network_schedule
        self._prio_epr = prio_epr

    def timebin_for_task(self, tid: int) -> EhiNetworkTimebin:
        assert self._task_graph is not None
        task = self._task_graph.get_tinfo(tid).task
        assert isinstance(task, SinglePairTask) or isinstance(task, MultiPairTask)
        drv_mem = self._driver._memory
        rrcall = drv_mem.read_shared_rrcall(task.shared_ptr)
        process = self._memmgr.get_process(task.pid)
        routine = process.get_request_routine(rrcall.routine_name)
        request = routine.request
        epr_sck = process.epr_sockets[request.epr_socket_id]
        return EhiNetworkTimebin(
            nodes=frozenset({self._node_id, epr_sck.remote_id}),
            pids={
                self._node_id: epr_sck.local_pid,
                epr_sck.remote_id: epr_sck.remote_pid,
            },
        )

    def are_resources_available(self, tid: int) -> bool:
        assert self._task_graph is not None
        task = self._task_graph.get_tinfo(tid).task
        self._task_logger.debug(f"check if resources available for task {tid} ({task})")
        phys_qubits_in_use = [
            i for i, vmap in self._memmgr._physical_mapping.items() if vmap is not None
        ]
        self._task_logger.debug(f"physical qubits in use: {phys_qubits_in_use}")
        if isinstance(task, SinglePairTask):
            # TODO: refactor
            drv_mem = self._driver._memory
            rrcall = drv_mem.read_shared_rrcall(task.shared_ptr)
            process = self._memmgr.get_process(task.pid)
            routine = process.get_request_routine(rrcall.routine_name)

            # Get virt ID which would be need to be allocated
            virt_id = routine.request.virt_ids.get_id(task.pair_index)

            # Check if virt ID is available by trying to allocate
            # (without actually allocating)

            try:
                self._memmgr.allocate(task.pid, virt_id)
                self._memmgr.free(task.pid, virt_id, send_signal=False)
                return True
            except AllocError:
                return False
        elif isinstance(task, MultiPairTask):
            # TODO: refactor
            drv_mem = self._driver._memory
            rrcall = drv_mem.read_shared_rrcall(task.shared_ptr)
            process = self._memmgr.get_process(task.pid)
            routine = process.get_request_routine(rrcall.routine_name)

            # Hack to get num_pairs (see comment in hostprocessor.py)
            prog_input = process.prog_instance.inputs.values
            if isinstance(routine.request.num_pairs, Template):
                template_name = routine.request.num_pairs.name
                num_pairs = prog_input[template_name]
            else:
                num_pairs = routine.request.num_pairs

            # Get virt IDs which would be need to be allocated
            if routine.request.virt_ids.typ == VirtIdMappingType.EQUAL:
                virt_id = routine.request.virt_ids.single_value
                assert virt_id is not None and isinstance(virt_id, int)
                virt_ids = [virt_id]
            else:
                virt_ids = [
                    routine.request.virt_ids.get_id(i) for i in range(num_pairs)
                ]

            # Check if virt IDs are available by trying to allocate
            # (without actually allocating)
            try:
                self._task_logger.debug(f"trying to allocate virt IDs {virt_ids}")
                temp_allocated: List[int] = []
                for virt_id in virt_ids:
                    self._memmgr.allocate(task.pid, virt_id)
                    temp_allocated.append(virt_id)  # successful alloc
                # Free all temporarily allocated qubits again
                for virt_id in temp_allocated:
                    self._memmgr.free(task.pid, virt_id, send_signal=False)
                self._task_logger.debug(f"all virt IDs available")
                return True
            except AllocError:
                # Make sure all qubits that did successfully allocate are freed
                for virt_id in temp_allocated:
                    self._memmgr.free(task.pid, virt_id, send_signal=False)
                self._task_logger.debug(f"some virt IDs unavailable")
                return False
        elif isinstance(task, LocalRoutineTask):
            drv_mem = self._driver._memory
            lrcall = drv_mem.read_shared_lrcall(task.shared_ptr)
            process = self._memmgr.get_process(task.pid)
            local_routine = process.get_local_routine(lrcall.routine_name)
            virt_ids = local_routine.metadata.qubit_use
            try:
                # get qubit IDs that are not already allocated
                new_ids = [
                    vid
                    for vid in virt_ids
                    if self._memmgr.phys_id_for(task.pid, vid) is None
                ]
                # try to allocate them
                temp_allocated: List[int] = []
                for virt_id in new_ids:
                    self._memmgr.allocate(task.pid, virt_id)
                    temp_allocated.append(virt_id)  # successful alloc
                # Free all temporarily allocated qubits again
                for virt_id in temp_allocated:
                    self._memmgr.free(task.pid, virt_id, send_signal=False)
                self._task_logger.debug(f"all virt IDs available")
                return True
            except AllocError:
                # Make sure all qubits that did successfully allocate are freed
                for virt_id in temp_allocated:
                    self._memmgr.free(task.pid, virt_id, send_signal=False)
                self._task_logger.debug(f"some virt IDs unavailable")
                return False
        else:
            self._logger.info(
                f"Checking if resources are available for task type {type(task)}, "
                "returning True but no actual check is implemented"
            )
            # NOTE: we assume that callback tasks never allocate any additional
            # resources so they can always return `True` here.
            return True

    def update_status(self) -> None:
        tg = self._task_graph

        if tg is None or len(tg.get_tasks()) == 0:
            self._status = SchedulerStatus(status={Status.GRAPH_EMPTY}, params={})
            return

        # All tasks that have no predecessors, internal nor external.
        no_predecessors = tg.get_roots()
        self._task_logger.debug(
            f"no_predecessors: {[str(tg.get_tinfo(t).task) for t in no_predecessors]}"
        )

        # All tasks that have only external predecessors.
        blocked_on_other_core = tg.get_tasks_blocked_only_on_external()
        self._task_logger.debug(
            f"blocked_on_other_core : {[str(tg.get_tinfo(t).task) for t in blocked_on_other_core]}"
        )

        # All EPR (SinglePair or MultiPair) tasks that have no predecessors,
        # internal nor external.
        epr_no_predecessors = [
            tid for tid in no_predecessors if tg.get_tinfo(tid).task.is_epr_task()
        ]
        self._task_logger.debug(
            f"epr_no_predecessors : {[str(tg.get_tinfo(t).task) for t in epr_no_predecessors]}"
        )

        # All tasks without predecessors for which not all resources are availables.
        blocked_on_resources = [
            tid for tid in no_predecessors if not self.are_resources_available(tid)
        ]
        self._task_logger.debug(
            f"blocked_on_resources : {[str(tg.get_tinfo(t).task) for t in blocked_on_resources]}"
        )

        # All non-EPR tasks that are ready for execution.
        non_epr_ready = [
            tid
            for tid in no_predecessors
            if tid not in epr_no_predecessors and tid not in blocked_on_resources
        ]
        self._task_logger.debug(
            f"non_epr_ready : {[str(tg.get_tinfo(t).task) for t in non_epr_ready]}"
        )

        # All EPR tasks that have no predecessors and are not blocked on resources.
        epr_no_preds_not_blocked = [
            tid for tid in epr_no_predecessors if tid not in blocked_on_resources
        ]
        self._task_logger.debug(
            f"epr_no_preds_not_blocked : {[str(tg.get_tinfo(t).task) for t in epr_no_preds_not_blocked]}"
        )

        # All EPR tasks that can be immediately executed.
        epr_ready = []

        # The next EPR task (if any) that is ready to execute but needs to wait for its
        # corresponding time bin.
        epr_wait_for_bin: Optional[Tuple[int, int]] = None  # (task ID, delta)

        time_until_bin: Dict[int, int] = {}  # task ID -> time until bin

        now = ns.sim_time()
        for e in epr_no_preds_not_blocked:
            if self._network_schedule is not None:
                # Find the time until the next netschedule timebin that allows this EPR task.
                bin = self.timebin_for_task(e)
                self._task_logger.info(f"EPR ready: task {e}, bin: {bin}")
                delta = self._network_schedule.next_specific_bin(now, bin)
                time_until_bin[e] = delta
                self._task_logger.info(f"EPR ready: task {e}, delta: {delta}")
                if delta == 0:
                    epr_ready.append(e)
            else:
                # No network schedule: immediate just execute the EPR task
                epr_ready.append(e)

        epr_non_zero_delta = {
            tid: delta for tid, delta in time_until_bin.items() if delta > 0
        }
        self._task_logger.info(f"epr_non_zero_delta: {epr_non_zero_delta}")
        if len(epr_non_zero_delta) > 0:
            sorted_by_delta = sorted(
                epr_non_zero_delta.items(), key=lambda item: item[1]
            )
            earliest, delta = sorted_by_delta[0]
            epr_wait_for_bin = (earliest, delta)

        self._task_logger.info(f"epr_wait_for_bin: {epr_wait_for_bin}")

        if len(epr_ready) > 0:
            self._task_logger.info(f"epr_ready: {epr_ready}")
            self._status = SchedulerStatus(
                status={Status.EPR_GEN}, params={"task_id": epr_ready[0]}
            )
        elif len(non_epr_ready) > 0:
            with_deadline = [
                t for t in non_epr_ready if tg.get_tinfo(t).deadline is not None
            ]

            if not self._use_deadlines:
                with_deadline = []
            if len(with_deadline) > 0:
                # Sort them by deadline and return the one with the earliest deadline
                deadlines = {t: tg.get_tinfo(t).deadline for t in with_deadline}
                sorted_by_deadline = sorted(deadlines.items(), key=lambda item: item[1])  # type: ignore
                to_return = sorted_by_deadline[0][0]
                self._logger.debug(f"Return task {to_return}")
                self._task_logger.debug(f"Return task {to_return}")
                self._status = SchedulerStatus(
                    status={Status.NEXT_TASK}, params={"task_id": to_return}
                )
            else:
                # No deadlines
                if self._deterministic:
                    index = 0
                else:
                    index = random.randint(0, len(non_epr_ready) - 1)
                to_return = non_epr_ready[index]
                self._logger.debug(f"Return task {to_return}")
                self._task_logger.debug(f"Return task {to_return}")
                self._status = SchedulerStatus(
                    status={Status.NEXT_TASK}, params={"task_id": to_return}
                )
        else:
            if len(blocked_on_other_core) > 0:
                self._logger.debug("Waiting other core")
                self._task_logger.debug("Waiting other core")
                self._status.status.add(Status.WAITING_OTHER_CORE)
            if len(blocked_on_resources) > 0:
                self._logger.debug("Waiting resources")
                self._task_logger.debug("Waiting resources")
                self._status.status.add(Status.WAITING_RESOURCES)
            if epr_wait_for_bin is not None:
                self._logger.debug("Waiting time bin")
                self._task_logger.debug("Waiting time bin")
                task_id, delta = epr_wait_for_bin
                self._status.status.add(Status.WAITING_TIME_BIN)
                self._status.params["delta"] = delta

            if len(self.status.status) == 0:
                raise RuntimeError

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            self._task_logger.debug("updating status...")
            self._status = SchedulerStatus(status=set(), params={})
            self.update_external_predcessors()
            self.update_status()
            self._task_logger.debug(f"status: {self.status.status}")
            if Status.EPR_GEN in self.status.status:
                task_id = self.status.params["task_id"]
                yield from self.handle_task(task_id)
            elif Status.NEXT_TASK in self.status.status:
                task_id = self.status.params["task_id"]
                yield from self.handle_task(task_id)
            else:
                ev_expr = self.await_port_input(self.node_scheduler_in_port)
                if Status.WAITING_OTHER_CORE in self.status.status:
                    ev_expr = ev_expr | self.await_signal(
                        sender=self._other_scheduler,
                        signal_label=SIGNAL_TASK_COMPLETED,
                    )
                if Status.WAITING_RESOURCES in self.status.status:
                    ev_expr = ev_expr | self.await_signal(
                        sender=self._memmgr,
                        signal_label=SIGNAL_MEMORY_FREED,
                    )
                if Status.WAITING_TIME_BIN in self.status.status:
                    delta = self.status.params["delta"]
                    self._schedule_after(delta, EVENT_WAIT)
                    ev_timebin = EventExpression(source=self, event_type=EVENT_WAIT)
                    ev_expr = ev_expr | ev_timebin
                self._task_logger.debug(f"yielding on {ev_expr}")
                yield ev_expr



from typing import List

import pytest
from netqasm.lang.instr import core, nv, trapped_ion
from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.common import MultiQubit
from qoala.lang.ehi import (
    EhiBuilder,
    EhiGateInfo,
    EhiLatencies,
    EhiLinkInfo,
    EhiNetworkInfo,
    EhiNodeInfo,
    EhiQubitInfo,
    UnitModule,
)


def qubit_info() -> EhiQubitInfo:
    return EhiQubitInfo(is_communication=True, decoherence_rate=0)


def single_gates() -> List[EhiGateInfo]:
    return [
        EhiGateInfo(instr, 5e3, 0)
        for instr in [
            core.InitInstruction,
            nv.RotXInstruction,
            nv.RotYInstruction,
            nv.RotZInstruction,
            core.MeasInstruction,
        ]
    ]


def multi_gates() -> List[EhiGateInfo]:
    return [EhiGateInfo(nv.ControlledRotXInstruction, 100e3, 0)]


def create_ehi() -> EhiNodeInfo:
    num_qubits = 3

    qubit_infos = {i: qubit_info() for i in range(num_qubits)}

    flavour = NVFlavour

    single_gate_infos = {i: single_gates() for i in range(num_qubits)}

    multi_gate_infos = {
        MultiQubit([0, 1]): multi_gates(),
        MultiQubit([1, 0]): multi_gates(),
    }

    all_qubit_gate_infos = [EhiGateInfo(trapped_ion.AllQubitsRotXInstruction, 50e3, 0)]

    return EhiNodeInfo(
        qubit_infos,
        flavour,
        single_gate_infos,
        multi_gate_infos,
        EhiLatencies.all_zero(),
        all_qubit_gate_infos,
    )


def test_1_qubit():
    ehi = create_ehi()

    for i in range(3):
        um = UnitModule.from_ehi(ehi, qubit_ids=[i])

        assert um.info.qubit_infos == {i: qubit_info()}
        assert um.info.single_gate_infos == {i: single_gates()}
        assert um.info.multi_gate_infos == {}
        assert um.info.all_qubit_gate_infos is None


def test_2_qubits():
    ehi = create_ehi()

    um01 = UnitModule.from_ehi(ehi, qubit_ids=[0, 1])

    # write test for error
    with pytest.raises(ValueError):
        um01 = UnitModule.from_ehi(ehi, qubit_ids=[0, 3])

    assert um01.info.qubit_infos == {0: qubit_info(), 1: qubit_info()}
    assert um01.info.single_gate_infos == {0: single_gates(), 1: single_gates()}
    assert um01.info.multi_gate_infos == {
        MultiQubit([0, 1]): multi_gates(),
        MultiQubit([1, 0]): multi_gates(),
    }
    assert um01.info.all_qubit_gate_infos is None

    um02 = UnitModule.from_ehi(ehi, qubit_ids=[0, 2])

    assert um02.info.qubit_infos == {0: qubit_info(), 2: qubit_info()}
    assert um02.info.single_gate_infos == {0: single_gates(), 2: single_gates()}
    assert um02.info.multi_gate_infos == {}
    assert um02.info.all_qubit_gate_infos is None

    um12 = UnitModule.from_ehi(ehi, qubit_ids=[1, 2])

    assert um12.info.qubit_infos == {1: qubit_info(), 2: qubit_info()}
    assert um12.info.single_gate_infos == {1: single_gates(), 2: single_gates()}
    assert um12.info.multi_gate_infos == {}
    assert um12.info.all_qubit_gate_infos is None


def test_full():
    ehi = create_ehi()
    um = UnitModule.from_full_ehi(ehi)

    assert um.info.qubit_infos == {i: qubit_info() for i in range(3)}
    assert um.info.single_gate_infos == {i: single_gates() for i in range(3)}
    assert um.info.multi_gate_infos == {
        MultiQubit([0, 1]): multi_gates(),
        MultiQubit([1, 0]): multi_gates(),
    }

    assert um.get_all_qubit_ids() == [0, 1, 2]


def test_perfect_qubit():
    qubit_info = EhiBuilder.perfect_qubit(is_communication=True)
    assert qubit_info.is_communication
    assert qubit_info.decoherence_rate == 0


def test_decoherence_qubit():
    qubit_info = EhiBuilder.decoherence_qubit(
        is_communication=True, decoherence_rate=0.3
    )
    assert qubit_info.is_communication
    assert qubit_info.decoherence_rate == 0.3


def test_perfect_gates():
    duration = 5e3
    instructions = [nv.RotXInstruction, nv.RotYInstruction]
    gate_infos = EhiBuilder.perfect_gates(duration, instructions)

    assert gate_infos == [
        EhiGateInfo(instruction=instr, duration=duration, decoherence=0)
        for instr in instructions
    ]


def test_decoherence_gates():
    duration = 5e3
    instructions = [nv.RotXInstruction, nv.RotYInstruction]
    gate_infos = EhiBuilder.decoherence_gates(duration, instructions, decoherence=0.2)

    assert gate_infos == [
        EhiGateInfo(instruction=instr, duration=duration, decoherence=0.2)
        for instr in instructions
    ]


def test_perfect_uniform():
    num_qubits = 3
    single_qubit_instructions = [nv.RotXInstruction, nv.RotYInstruction]
    single_gate_duration = 5e3
    two_qubit_instructions = [nv.ControlledRotXInstruction]
    two_gate_duration = 2e5
    all_qubit_instructions = [trapped_ion.AllQubitsRotXInstruction]
    all_qubit_gate_duration = 1e5
    ehi = EhiBuilder.perfect_uniform(
        num_qubits,
        NVFlavour,
        single_qubit_instructions,
        single_gate_duration,
        two_qubit_instructions,
        two_gate_duration,
        all_qubit_instructions,
        all_qubit_gate_duration,
    )

    assert ehi.qubit_infos == {
        i: EhiQubitInfo(is_communication=True, decoherence_rate=0)
        for i in range(num_qubits)
    }

    assert ehi.single_gate_infos == {
        i: [
            EhiGateInfo(instruction=instr, duration=single_gate_duration, decoherence=0)
            for instr in single_qubit_instructions
        ]
        for i in range(num_qubits)
    }

    assert ehi.multi_gate_infos == {
        MultiQubit([i, j]): [
            EhiGateInfo(instruction=instr, duration=two_gate_duration, decoherence=0)
            for instr in two_qubit_instructions
        ]
        for i in range(num_qubits)
        for j in range(num_qubits)
        if i != j
    }

    assert ehi.all_qubit_gate_infos == [
        EhiGateInfo(trapped_ion.AllQubitsRotXInstruction, all_qubit_gate_duration, 0)
    ]


def test_build_fully_uniform():
    qubit_info = EhiQubitInfo(is_communication=True, decoherence_rate=0.01)
    single_gate_infos = [
        EhiGateInfo(instruction=instr, duration=5e3, decoherence=0.02)
        for instr in [nv.RotXInstruction, nv.RotYInstruction]
    ]
    two_gate_infos = [
        EhiGateInfo(
            instruction=nv.ControlledRotXInstruction, duration=2e4, decoherence=0.05
        )
    ]
    all_qubit_gate_infos = [EhiGateInfo(trapped_ion.AllQubitsRotXInstruction, 1e5, 0)]

    ehi = EhiBuilder.fully_uniform(
        num_qubits=3,
        flavour=NVFlavour,
        qubit_info=qubit_info,
        single_gate_infos=single_gate_infos,
        two_gate_infos=two_gate_infos,
        all_qubit_gate_infos=all_qubit_gate_infos,
    )

    assert len(ehi.qubit_infos) == 3
    for i in range(3):
        assert ehi.qubit_infos[i].is_communication
        single_gates = [info.instruction for info in ehi.single_gate_infos[i]]
        assert nv.RotXInstruction in single_gates
        assert nv.RotYInstruction in single_gates

    for i in range(3):
        for j in range(3):
            if i == j:
                continue
            multi = MultiQubit([i, j])
            gates = [info.instruction for info in ehi.multi_gate_infos[multi]]
            assert nv.ControlledRotXInstruction in gates

    assert ehi.all_qubit_gate_infos == all_qubit_gate_infos


def test_perfect_star():
    num_qubits = 3
    comm_instructions = [nv.RotXInstruction, nv.RotYInstruction, nv.RotZInstruction]
    comm_duration = 5e3
    mem_instructions = [nv.RotXInstruction, nv.RotYInstruction]
    mem_duration = 1e4
    two_instructions = [nv.ControlledRotXInstruction, nv.ControlledRotYInstruction]
    two_duration = 2e5
    ehi = EhiBuilder.perfect_star(
        num_qubits,
        NVFlavour,
        comm_instructions,
        comm_duration,
        mem_instructions,
        mem_duration,
        two_instructions,
        two_duration,
    )

    assert ehi.qubit_infos[0] == EhiBuilder.perfect_qubit(is_communication=True)

    for i in range(1, num_qubits):
        assert ehi.qubit_infos[i] == EhiBuilder.perfect_qubit(is_communication=False)

    assert ehi.single_gate_infos[0] == EhiBuilder.perfect_gates(
        comm_duration, [nv.RotXInstruction, nv.RotYInstruction, nv.RotZInstruction]
    )
    for i in range(1, num_qubits):
        assert ehi.single_gate_infos[i] == EhiBuilder.perfect_gates(
            mem_duration, [nv.RotXInstruction, nv.RotYInstruction]
        )

    for i in range(1, num_qubits):
        assert ehi.multi_gate_infos[MultiQubit([0, i])] == EhiBuilder.perfect_gates(
            two_duration, [nv.ControlledRotXInstruction, nv.ControlledRotYInstruction]
        )


def test_generic_t1t2_star():
    num_qubits = 3
    comm_decoherence = 0.05
    mem_decoherence = 0.01
    comm_instructions = [nv.RotXInstruction, nv.RotYInstruction, nv.RotZInstruction]
    comm_duration = 5e3
    comm_instr_decoherence = 0.3
    mem_instructions = [nv.RotXInstruction, nv.RotYInstruction]
    mem_duration = 1e4
    mem_instr_decoherence = 0.4
    two_instructions = [nv.ControlledRotXInstruction, nv.ControlledRotYInstruction]
    two_duration = 2e5
    two_instr_decoherence = 0.5
    ehi = EhiBuilder.generic_t1t2_star(
        num_qubits,
        NVFlavour,
        comm_decoherence,
        mem_decoherence,
        comm_instructions,
        comm_duration,
        comm_instr_decoherence,
        mem_instructions,
        mem_duration,
        mem_instr_decoherence,
        two_instructions,
        two_duration,
        two_instr_decoherence,
    )

    assert ehi.qubit_infos[0] == EhiBuilder.decoherence_qubit(True, comm_decoherence)

    for i in range(1, num_qubits):
        assert ehi.qubit_infos[i] == EhiBuilder.decoherence_qubit(
            False, mem_decoherence
        )

    assert ehi.single_gate_infos[0] == EhiBuilder.decoherence_gates(
        comm_duration,
        [nv.RotXInstruction, nv.RotYInstruction, nv.RotZInstruction],
        comm_instr_decoherence,
    )
    for i in range(1, num_qubits):
        assert ehi.single_gate_infos[i] == EhiBuilder.decoherence_gates(
            mem_duration,
            [nv.RotXInstruction, nv.RotYInstruction],
            mem_instr_decoherence,
        )

    for i in range(1, num_qubits):
        assert ehi.multi_gate_infos[MultiQubit([0, i])] == EhiBuilder.decoherence_gates(
            two_duration,
            [nv.ControlledRotXInstruction, nv.ControlledRotYInstruction],
            two_instr_decoherence,
        )


def test_find_gates():
    num_qubits = 3
    comm_decoherence = 0.05
    mem_decoherence = 0.01
    comm_instructions = [nv.RotXInstruction, nv.RotYInstruction, nv.RotZInstruction]
    comm_duration = 5e3
    comm_instr_decoherence = 0.3
    mem_instructions = [nv.RotXInstruction, nv.RotYInstruction]
    mem_duration = 1e4
    mem_instr_decoherence = 0.4
    two_instructions = [nv.ControlledRotXInstruction, nv.ControlledRotYInstruction]
    two_duration = 2e5
    two_instr_decoherence = 0.5
    ehi = EhiBuilder.generic_t1t2_star(
        num_qubits,
        NVFlavour,
        comm_decoherence,
        mem_decoherence,
        comm_instructions,
        comm_duration,
        comm_instr_decoherence,
        mem_instructions,
        mem_duration,
        mem_instr_decoherence,
        two_instructions,
        two_duration,
        two_instr_decoherence,
    )
    for i in range(3):
        assert ehi.find_single_gate(i, nv.RotXInstruction) is not None
        assert ehi.find_single_gate(i, nv.RotYInstruction) is not None
        if i == 0:
            assert ehi.find_single_gate(i, nv.RotZInstruction) is not None
        else:
            assert ehi.find_single_gate(i, nv.RotZInstruction) is None

    for i in range(1, 3):
        assert ehi.find_multi_gate([0, i], nv.ControlledRotXInstruction) is not None
        assert ehi.find_multi_gate([0, i], nv.ControlledRotYInstruction) is not None
        assert ehi.find_multi_gate([i, 0], nv.ControlledRotYInstruction) is None


def test_find_all_qubit_gates():
    qubit_info = EhiQubitInfo(is_communication=True, decoherence_rate=0.01)
    all_qubit_gate_infos = [
        EhiGateInfo(trapped_ion.AllQubitsRotXInstruction, 1e5, 0),
        EhiGateInfo(trapped_ion.AllQubitsRotYInstruction, 1e5, 0),
        EhiGateInfo(trapped_ion.AllQubitsRotZInstruction, 1e5, 0),
    ]

    ehi = EhiBuilder.fully_uniform(
        num_qubits=3,
        flavour=NVFlavour,
        qubit_info=qubit_info,
        single_gate_infos=[],
        two_gate_infos=[],
        all_qubit_gate_infos=all_qubit_gate_infos,
    )
    assert ehi.find_all_qubit_gate(trapped_ion.AllQubitsRotXInstruction) is not None
    assert ehi.find_all_qubit_gate(trapped_ion.AllQubitsRotYInstruction) is not None
    assert ehi.find_all_qubit_gate(trapped_ion.AllQubitsRotZInstruction) is not None


def test_network_ehi():
    nodes = {0: "node0", 1: "node1", 2: "node2"}
    network_ehi = EhiNetworkInfo.perfect_fully_connected(nodes=nodes, duration=1000)

    assert network_ehi.get_link(0, 1) == EhiLinkInfo(duration=1000, fidelity=1.0)
    assert network_ehi.get_link(0, 2) == EhiLinkInfo(duration=1000, fidelity=1.0)
    assert network_ehi.get_link(1, 2) == EhiLinkInfo(duration=1000, fidelity=1.0)

    assert network_ehi.get_all_node_names() == ["node0", "node1", "node2"]

    assert network_ehi.get_node_id("node0") == 0
    assert network_ehi.get_node_id("node1") == 1
    assert network_ehi.get_node_id("node2") == 2


def test_network_ehi2():
    nodes = {0: "node0", 1: "node1", 2: "node2"}
    link_info = EhiLinkInfo(duration=1000, fidelity=0.9)
    network_ehi = EhiNetworkInfo.fully_connected(nodes=nodes, info=link_info)

    assert network_ehi.get_link(0, 1) == link_info
    assert network_ehi.get_link(0, 2) == link_info
    assert network_ehi.get_link(1, 2) == link_info

    assert network_ehi.get_all_node_names() == ["node0", "node1", "node2"]

    assert network_ehi.get_node_id("node0") == 0
    assert network_ehi.get_node_id("node1") == 1
    assert network_ehi.get_node_id("node2") == 2


def test_network_ehi3():
    nodes = {0: "node0", 1: "node1", 2: "node2"}
    network_ehi = EhiNetworkInfo.only_nodes(nodes=nodes)

    with pytest.raises(ValueError):
        network_ehi.get_link(0, 1)

    link_info = EhiLinkInfo(duration=1000, fidelity=0.7)
    network_ehi.add_link(0, 1, link_info)

    assert network_ehi.get_link(0, 1) == link_info

    # Already existing link (order does not matter)
    with pytest.raises(ValueError):
        network_ehi.add_link(1, 0, link_info)

    # Cannot create link with itself
    with pytest.raises(ValueError):
        network_ehi.add_link(0, 0, link_info)

    # Cannot create link with non-existing node
    with pytest.raises(ValueError):
        network_ehi.add_link(0, 4, link_info)


if __name__ == "__main__":
    test_1_qubit()
    test_2_qubits()
    test_full()
    test_perfect_qubit()
    test_decoherence_qubit()
    test_perfect_gates()
    test_decoherence_gates()
    test_perfect_uniform()
    test_build_fully_uniform()
    test_perfect_star()
    test_generic_t1t2_star()
    test_find_gates()
    test_find_all_qubit_gates()
    test_network_ehi()
    test_network_ehi2()
    test_network_ehi3()



from qoala.sim.qnos.qnos import Qnos
from qoala.sim.qnos.qnoscomp import QnosComponent
from qoala.sim.qnos.qnosinterface import QnosInterface, QnosLatencies
from qoala.sim.qnos.qnosprocessor import GenericProcessor, NVProcessor, QnosProcessor



from typing import List, Optional, Tuple

import pytest

from qoala.runtime.task import (
    HostEventTask,
    HostLocalTask,
    LocalRoutineTask,
    MultiPairCallbackTask,
    MultiPairTask,
    PreCallTask,
    ProcessorType,
    QoalaTask,
    TaskGraph,
    TaskGraphBuilder,
)


class SimpleTask(QoalaTask):
    def __init__(self, task_id: int) -> None:
        super().__init__(task_id, ProcessorType.CPU, 0)


def test_linear():
    tasks = [SimpleTask(i) for i in range(5)]
    precedences = [(i - 1, i) for i in range(1, 5)]
    rel_deadlines = [((i - 1, i), 100) for i in range(1, 5)]
    graph = TaskGraph()
    graph.add_tasks(tasks)
    graph.add_precedences(precedences)
    graph.add_rel_deadlines(rel_deadlines)

    assert graph.get_roots() == [0]
    assert graph.get_tinfo(0).predecessors == set()
    assert graph.get_tinfo(4).successors == set()
    assert all(graph.get_tinfo(i).predecessors == {i - 1} for i in range(1, 5))
    assert all(graph.get_tinfo(i).successors == {i + 1} for i in range(0, 4))
    assert all(graph.get_tinfo(i).deadline is None for i in range(5))
    assert graph.get_tinfo(0).rel_deadlines == {}
    assert all(graph.get_tinfo(i).rel_deadlines == {i - 1: 100} for i in range(1, 5))
    assert len(graph.get_tasks()) == 5

    graph.remove_task(0)

    with pytest.raises(AssertionError):
        graph.get_tinfo(0)

    assert graph.get_roots() == [1]
    assert graph.get_tinfo(1).predecessors == set()
    assert graph.get_tinfo(4).successors == set()
    assert all(graph.get_tinfo(i).predecessors == {i - 1} for i in range(2, 5))
    assert all(graph.get_tinfo(i).successors == {i + 1} for i in range(1, 4))
    assert graph.get_tinfo(1).deadline == 100
    assert all(graph.get_tinfo(i).deadline is None for i in range(2, 5))
    assert graph.get_tinfo(1).rel_deadlines == {}
    assert all(graph.get_tinfo(i).rel_deadlines == {i - 1: 100} for i in range(2, 5))
    assert len(graph.get_tasks()) == 4

    with pytest.raises(AssertionError):
        # not a root
        graph.remove_task(4)


def test_no_precedence():
    tasks = [SimpleTask(i) for i in range(5)]
    rel_deadlines = [((i - 1, i), 100) for i in range(1, 5)]
    graph = TaskGraph()
    graph.add_tasks(tasks)
    graph.add_rel_deadlines(rel_deadlines)

    assert graph.get_roots() == [i for i in range(5)]
    assert all(graph.get_tinfo(i).predecessors == set() for i in range(5))
    assert all(graph.get_tinfo(i).successors == set() for i in range(5))
    assert all(graph.get_tinfo(i).deadline is None for i in range(5))
    assert graph.get_tinfo(0).rel_deadlines == {}
    assert all(graph.get_tinfo(i).rel_deadlines == {i - 1: 100} for i in range(1, 5))

    graph.remove_task(0)

    with pytest.raises(AssertionError):
        graph.get_tinfo(0)

    assert graph.get_roots() == [i for i in range(1, 5)]
    assert all(graph.get_tinfo(i).predecessors == set() for i in range(1, 5))
    assert all(graph.get_tinfo(i).successors == set() for i in range(1, 5))
    assert graph.get_tinfo(1).deadline == 100
    assert all(graph.get_tinfo(i).deadline is None for i in range(2, 5))
    assert graph.get_tinfo(1).rel_deadlines == {}
    assert all(graph.get_tinfo(i).rel_deadlines == {i - 1: 100} for i in range(2, 5))

    graph.remove_task(4)

    with pytest.raises(AssertionError):
        graph.get_tinfo(4)

    assert graph.get_roots() == [1, 2, 3]
    assert all(graph.get_tinfo(i).predecessors == set() for i in range(1, 4))
    assert all(graph.get_tinfo(i).rel_deadlines == {i - 1: 100} for i in range(2, 4))


def test_get_partial_graph():
    pid = 0
    mp_ptr = 0
    lr_ptr = 1

    hl1 = HostLocalTask(0, pid, "hl1")
    hl2 = HostLocalTask(1, pid, "hl2")
    hl3 = HostLocalTask(2, pid, "hl2")
    he1 = HostEventTask(3, pid, "he1")
    prc1 = PreCallTask(4, pid, "prc1", mp_ptr)
    prc2 = PreCallTask(5, pid, "prc2", lr_ptr)
    poc1 = PreCallTask(6, pid, "poc1", mp_ptr)
    poc2 = PreCallTask(7, pid, "poc2", lr_ptr)
    mp1 = MultiPairTask(8, pid, mp_ptr)
    mpc1 = MultiPairCallbackTask(9, pid, "mpc1", mp_ptr)
    lr1 = LocalRoutineTask(10, pid, "lr1", lr_ptr)

    precedences = [
        (hl1.task_id, hl2.task_id),
        (hl1.task_id, he1.task_id),
        (hl2.task_id, prc1.task_id),
        (he1.task_id, prc1.task_id),
        (he1.task_id, prc2.task_id),
        (prc1.task_id, mp1.task_id),
        (mp1.task_id, mpc1.task_id),
        (mpc1.task_id, poc1.task_id),
        (prc2.task_id, lr1.task_id),
        (mpc1.task_id, lr1.task_id),
        (lr1.task_id, poc2.task_id),
        (poc1.task_id, hl3.task_id),
        (poc2.task_id, hl3.task_id),
    ]
    graph = TaskGraph()
    graph.add_tasks([hl1, hl2, hl3, he1, prc1, prc2, poc1, poc2, mp1, mpc1, lr1])
    graph.add_precedences(precedences)

    # Test immediate cross-predecessors
    for task in [hl1, hl2, he1, prc1, prc2, mpc1, hl3]:
        assert graph.cross_predecessors(task.task_id) == set()
    assert graph.cross_predecessors(mp1.task_id) == {prc1.task_id}
    assert graph.cross_predecessors(poc1.task_id) == {mpc1.task_id}
    assert graph.cross_predecessors(lr1.task_id) == {prc2.task_id}
    assert graph.cross_predecessors(poc2.task_id) == {lr1.task_id}

    # Test indirect cross-predecessors
    for task in [hl1, hl2, he1, prc1, prc2]:
        assert graph.cross_predecessors(task.task_id, immediate=False) == set()
    assert graph.cross_predecessors(mp1.task_id, immediate=False) == {prc1.task_id}
    assert graph.cross_predecessors(mpc1.task_id, immediate=False) == {prc1.task_id}
    assert graph.cross_predecessors(poc1.task_id, immediate=False) == {mpc1.task_id}
    assert graph.cross_predecessors(lr1.task_id, immediate=False) == {
        prc1.task_id,
        prc2.task_id,
    }
    assert graph.cross_predecessors(poc2.task_id, immediate=False) == {lr1.task_id}
    assert graph.cross_predecessors(hl3.task_id, immediate=False) == {
        mpc1.task_id,
        lr1.task_id,
    }

    assert all(
        graph.double_cross_predecessors(t.task_id) == set()
        for t in [hl1, hl2, hl3, he1, prc1, prc2, mp1, mpc1, lr1]
    )
    assert graph.double_cross_predecessors(poc1.task_id) == {prc1.task_id}
    assert graph.double_cross_predecessors(poc2.task_id) == {prc1.task_id, prc2.task_id}

    # Check CPU graph
    expected_cpu_precedences = [
        (hl1.task_id, hl2.task_id),
        (hl1.task_id, he1.task_id),
        (hl2.task_id, prc1.task_id),
        (he1.task_id, prc1.task_id),
        (he1.task_id, prc2.task_id),
        (prc1.task_id, poc1.task_id),
        (prc1.task_id, poc2.task_id),
        (prc2.task_id, poc2.task_id),
        (poc1.task_id, hl3.task_id),
        (poc2.task_id, hl3.task_id),
    ]
    expected_external_cpu_precedences = [
        (mpc1.task_id, poc1.task_id),
        (lr1.task_id, poc2.task_id),
    ]
    expected_cpu_graph = TaskGraph()
    expected_cpu_graph.add_tasks([hl1, hl2, hl3, he1, prc1, prc2, poc1, poc2])
    expected_cpu_graph.add_precedences(expected_cpu_precedences)
    expected_cpu_graph.add_ext_precedences(expected_external_cpu_precedences)
    cpu_graph = graph.get_cpu_graph()
    assert cpu_graph == expected_cpu_graph

    # Check QPU graph
    expected_qpu_precedences = [
        (mp1.task_id, mpc1.task_id),
        (mpc1.task_id, lr1.task_id),
    ]
    expected_external_qpu_precedences = [
        (prc1.task_id, mp1.task_id),
        (prc2.task_id, lr1.task_id),
    ]
    qpu_graph = graph.get_qpu_graph()
    expected_qpu_graph = TaskGraph()
    expected_qpu_graph.add_tasks([mp1, mpc1, lr1])
    expected_qpu_graph.add_precedences(expected_qpu_precedences)
    expected_qpu_graph.add_ext_precedences(expected_external_qpu_precedences)
    assert qpu_graph == expected_qpu_graph


def test_dynamic_update():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0), SimpleTask(1)])

    # task 0 should start at <2000 from now
    graph.add_deadlines([(0, 2000)])
    # task 1 should start <100 after task 1 finishes
    graph.add_rel_deadlines([((0, 1), 100)])

    # Mock execution of task 0, taking 500 time units.
    # First decrease all current absolute deadlines since removing task 0 will make
    # the relative deadline of task 1 an absolute deadline, which we do not want to decrease.
    graph.decrease_deadlines(500)
    graph.remove_task(0)
    assert graph.get_tinfo(1).deadline == 100


def test_linear_tasks():
    tasks = [SimpleTask(0), SimpleTask(1), SimpleTask(2), SimpleTask(3), SimpleTask(4)]

    graph = TaskGraphBuilder.linear_tasks(tasks)
    assert graph.get_tinfo(0).task == tasks[0]
    assert graph.get_tinfo(0).predecessors == set()
    assert graph.get_tinfo(len(tasks) - 1).successors == set()
    for i in range(len(tasks)):
        assert graph.get_tinfo(i).task == tasks[i]
        if i > 0:
            assert graph.get_tinfo(i).predecessors == {i - 1}
        if i < len(tasks) - 1:
            assert graph.get_tinfo(i).successors == {i + 1}


def test_linear_tasks_with_timestamps():
    tasks = [SimpleTask(0), SimpleTask(1), SimpleTask(2), SimpleTask(3), SimpleTask(4)]

    start_times: List[Tuple[SimpleTask, Optional[int]]] = [
        (tasks[0], 0),
        (tasks[1], 2000),
        (tasks[2], 3000),
        (tasks[3], 12500),
        (tasks[4], None),
    ]

    graph = TaskGraphBuilder.linear_tasks_with_start_times(start_times)
    assert graph.get_tinfo(0).task == tasks[0]
    assert graph.get_tinfo(0).predecessors == set()
    assert graph.get_tinfo(len(tasks) - 1).successors == set()
    for i in range(len(tasks)):
        assert graph.get_tinfo(i).task == tasks[i]
        if i > 0:
            assert graph.get_tinfo(i).predecessors == {i - 1}
        if i < len(tasks) - 1:
            assert graph.get_tinfo(i).successors == {i + 1}

    for task, start_time in start_times:
        assert graph.get_tinfo(task.task_id).start_time == start_time


def test_merge():
    graph1 = TaskGraph()
    graph1.add_tasks([SimpleTask(0), SimpleTask(1)])
    graph1.add_precedences([(0, 1)])

    graph2 = TaskGraph()
    graph2.add_tasks([SimpleTask(2), SimpleTask(3)])
    graph2.add_precedences([(2, 3)])

    merged = TaskGraphBuilder.merge([graph1, graph2])
    for i in range(4):
        assert merged.get_tinfo(i).task == SimpleTask(i)

    assert merged.get_tinfo(1).predecessors == {0}
    assert merged.get_tinfo(0).successors == {1}
    assert merged.get_tinfo(3).predecessors == {2}
    assert merged.get_tinfo(2).successors == {3}


def test_merge_linear():
    graph1 = TaskGraph()
    graph1.add_tasks([SimpleTask(0), SimpleTask(1)])
    graph1.add_precedences([(0, 1)])

    graph2 = TaskGraph()
    graph2.add_tasks([SimpleTask(2), SimpleTask(3)])
    graph2.add_precedences([(2, 3)])

    merged = TaskGraphBuilder.merge_linear([graph1, graph2])
    for i in range(4):
        assert merged.get_tinfo(i).task == SimpleTask(i)

    assert merged.get_tinfo(1).predecessors == {0}
    assert merged.get_tinfo(0).successors == {1}
    assert merged.get_tinfo(3).predecessors == {2}
    assert merged.get_tinfo(2).successors == {3}

    # Check that there is precedence between two original graphs
    assert merged.get_tinfo(2).predecessors == {1}
    assert merged.get_tinfo(1).successors == {2}


def test_linearize_1():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0), SimpleTask(1)])
    with pytest.raises(RuntimeError):
        graph.linearize()


def test_linearize_2():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0), SimpleTask(1)])
    graph.add_precedences([(0, 1)])
    assert graph.linearize() == [0, 1]


def test_linearize_3():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0), SimpleTask(1), SimpleTask(2)])
    graph.add_precedences([(0, 1)])
    graph.add_precedences([(0, 2)])
    with pytest.raises(RuntimeError):
        graph.linearize()


def test_linearize_4():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0), SimpleTask(1), SimpleTask(2)])
    graph.add_precedences([(1, 2)])
    graph.add_precedences([(2, 0)])
    assert graph.linearize() == [1, 2, 0]


if __name__ == "__main__":
    test_linear()
    test_no_precedence()
    test_get_partial_graph()
    test_dynamic_update()
    test_linear_tasks()
    test_linear_tasks_with_timestamps()
    test_merge()
    test_merge_linear()
    test_linearize_1()
    test_linearize_2()
    test_linearize_3()
    test_linearize_4()



from __future__ import annotations

import os
from dataclasses import dataclass

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app_separate_inputs


def create_procnode_cfg(name: str, id: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(1),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class ClassicalAppResult:
    alice_results: BatchResult
    bob_results: BatchResult


def run_classical_app(num_iterations: int) -> ClassicalAppResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg("alice", alice_id)
    bob_node_cfg = create_procnode_cfg("bob", bob_id)

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, 2500)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn]

    alice_program = load_program("classical_alice.iqoala")
    alice_inputs = [
        ProgramInput({"bob_id": bob_id, "value": i}) for i in range(num_iterations)
    ]

    bob_program = load_program("classical_bob.iqoala")
    bob_inputs = [ProgramInput({"alice_id": alice_id}) for _ in range(num_iterations)]

    app_result = run_two_node_app_separate_inputs(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_inputs, "bob": bob_inputs},
        network_cfg=network_cfg,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return ClassicalAppResult(alice_result, bob_result)


def test_classical_app():
    # LogManager.set_log_level("INFO")
    # LogManager.enable_task_logger(True)
    # LogManager.log_tasks_to_file("test_classical.log")
    num_iterations = 20

    result = run_classical_app(num_iterations=num_iterations)

    program_results = result.alice_results.results
    outcomes = [result.values["returned_value"] for result in program_results]
    print(outcomes)
    # assert all(outcome == 1 for outcome in outcomes)


if __name__ == "__main__":
    test_classical_app()



import functools as ft
from typing import List, Tuple

import numpy as np
from netsquid.components.instructions import Instruction
from netsquid.components.qmemory import QuantumMemory
from netsquid.qubits import operators as ops
from scipy.linalg import expm


class IMeasAll(Instruction):
    """
    Measurement instruction for all qubits in a quantum memory.
    """

    @property
    def name(self) -> str:
        """instruction name."""
        return "measure_all"

    @property
    def num_positions(self) -> int:
        """number of targeted memory positions. If -1, number is unrestricted."""
        return -1

    def execute(
        self, quantum_memory: QuantumMemory, positions: List[int], **kwargs
    ) -> List[List[int]]:
        """Execute instruction on a quantum memory.

        :param quantum_memory: NetSquid Quantum memory to execute instruction on.
        :param positions: Memory positions_of_connections to execute instruction on.
        """
        result = []
        for pos in positions:
            result.append(quantum_memory.measure(positions=[pos])[0][0])
        return [result]


class IRotationAllGate(Instruction):
    """
    Rotation instruction for all qubits in a quantum memory.
    """

    def __init__(self, name: str, axis: Tuple[int, int, int]) -> None:
        self._name = name
        self._axis = axis

    @property
    def name(self) -> str:
        """instruction name."""
        return self._name

    @property
    def num_positions(self) -> int:
        """number of targeted memory positions. If -1, number is unrestricted."""
        return -1

    def execute(
        self,
        quantum_memory: QuantumMemory,
        positions: List[int],
        angle: float,
        **kwargs
    ) -> None:
        """Execute instruction on a quantum memory.

        :param quantum_memory: NetSquid Quantum memory to execute instruction on.
        :param positions: Memory positions_of_connections to execute instruction on.
        """
        # print("executing rotation instruction with angle: ", angle)
        operator = ops.create_rotation_op(angle=angle, rotation_axis=self._axis)
        for pos in positions:
            quantum_memory.operate(positions=pos, operator=operator)


class IBichromaticGate(Instruction):
    """
    Bichromatic gate instruction.
    """

    @property
    def name(self) -> str:
        """instruction name."""
        return "bichromatic"

    @property
    def num_positions(self) -> int:
        """number of targeted memory positions. If -1, number is unrestricted."""
        return -1

    def construct_operator(self, n: int, angle: float) -> ops.Operator:
        """Construct operator which is applied by the gate.
        Used by execute method.

        :param n: Number of qubits to apply operator on.
        :param angle: Angle of rotation.


        In case of n=2, it constructs the XX gate. Specifically, the matrix:
        [ cos(angle/2)    0                0               -i sin(angle/2) ]
        [ 0               cos(angle/2)     -i*sin(angle/2)  0              ]
        [ 0               -i*sin(angle/2)  cos(angle/2)     0              ]
        [ -i*sin(angle/2)    0             0                cos(angle/2)   ]

        Note that here "angle/2" is used, in contrast to
        https://arxiv.org/pdf/1603.07678.pdf and
        https://ionq.com/docs/getting-started-with-native-gates#entangling-gates !

        I.e. to get an XX(t) gate as specified in the above paper, use angle = 2*t
        for this BichromaticGate defined here.

        For two qubits, this Bichromatic gates may be used to do a CNOT as follows:

        NETQASM:
        // cnot between q0 and q1
        rot_x_all 8 4
        rot_z Q0 8 4
        rot_x_all 24 4
        bichromatic 8 4
        rot_x_all 24 4
        rot_x_all 8 4
        rot_z Q0 24 4
        rot_x_all 24 4
        """
        sub_matrices = []
        X = np.array([[0, 1], [1, 0]])
        for i in range(2, n + 1):
            for j in range(1, i):
                identity1 = np.eye(2 ** (j - 1))
                identity2 = np.eye(2 ** (i - j - 1))
                identity3 = np.eye(2 ** (n - i))
                sub_matrix_parts = [identity1, X, identity2, X, identity3]
                sub_matrix = ft.reduce(np.kron, sub_matrix_parts)
                sub_matrices.append(sub_matrix)

        matrix = expm(-1j * angle / 2 * np.sum(sub_matrices, axis=0))
        operator = ops.Operator(name="bichromatic_operator", matrix=matrix)
        return operator

    def execute(
        self,
        quantum_memory: QuantumMemory,
        positions: List[int],
        angle: float,
        **kwargs
    ) -> None:
        """Execute instruction on a quantum memory.

        :param quantum_memory: NetSquid Quantum memory to execute instruction on.
        :param positions: Memory positions to do instruction on. Can be empty.
        """

        operator = self.construct_operator(n=len(positions), angle=angle)
        # print("operator: \n", np.around(operator.arr, 2))
        quantum_memory.operate(positions=positions, operator=operator)


INSTR_MEASURE_ALL = IMeasAll()
INSTR_ROT_X_ALL = IRotationAllGate("x_rot_all_gate", axis=(1, 0, 0))
INSTR_ROT_Y_ALL = IRotationAllGate("y_rot_all_gate", axis=(0, 1, 0))
INSTR_ROT_Z_ALL = IRotationAllGate("z_rot_all_gate", axis=(0, 0, 1))
INSTR_BICHROMATIC = IBichromaticGate()



from __future__ import annotations

import os
from dataclasses import dataclass

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.from_nv_params(
            num_qubits=num_qubits, params=NvParams()
        ),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=100_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("NvNtf"),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text, flavour=NVFlavour()).parse()


@dataclass
class PingPongResult:
    alice_result: BatchResult
    bob_result: BatchResult


def run_pingpong(num_iterations: int) -> PingPongResult:
    ns.sim_reset()

    num_qubits = 3
    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=500_000
    )

    alice_program = load_program("pingpong_nv_alice.iqoala")
    alice_input = ProgramInput({"bob_id": bob_id})

    bob_program = load_program("pingpong_nv_bob.iqoala")
    bob_input = ProgramInput({"alice_id": alice_id})

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_input, "bob": bob_input},
        network_cfg=network_cfg,
        linear=True,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return PingPongResult(alice_result, bob_result)


def check_pingpong(num_iterations: int):
    # LogManager.set_log_level("DEBUG")
    # LogManager.log_to_file("pingpong_nv.log")

    ns.sim_reset()
    result = run_pingpong(num_iterations=num_iterations)

    program_results = result.alice_result.results
    outcomes = [result.values["outcome"] for result in program_results]
    print(outcomes)
    assert all(outcome == 1 for outcome in outcomes)


def test_pingpong():
    check_pingpong(10)


if __name__ == "__main__":
    test_pingpong()



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import List

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app_separate_inputs


def create_procnode_cfg(name: str, id: int, determ: bool) -> ProcNodeConfig:
    nv_params = NvParams()
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.from_nv_params(num_qubits=5, params=nv_params),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("NvNtf"),
        determ_sched=determ,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text, flavour=NVFlavour()).parse()


@dataclass
class RspResult:
    alice_results: BatchResult
    bob_results: BatchResult


def run_rsp(num_iterations: int, angles: List[int]) -> RspResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg("alice", alice_id, determ=True)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, determ=True)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )

    alice_program = load_program("rsp_alice.iqoala")
    bob_program = load_program("rsp_bob.iqoala")

    alice_inputs = [
        ProgramInput({"bob_id": bob_id, "angle": angle}) for angle in angles
    ]
    bob_inputs = [ProgramInput({"alice_id": alice_id}) for _ in range(num_iterations)]

    app_result = run_two_node_app_separate_inputs(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_inputs, "bob": bob_inputs},
        network_cfg=network_cfg,
        linear=True,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return RspResult(alice_result, bob_result)


def test_rsp():
    # LogManager.set_log_level("DEBUG")
    # LogManager.log_tasks_to_file("teleport_plus_local.log")
    num_iterations = 20

    angles = [16 for _ in range(num_iterations)]

    result = run_rsp(num_iterations, angles)
    program_results = result.bob_results.results
    print(program_results)
    outcomes = [result.values["outcome"] for result in program_results]
    print(outcomes)


if __name__ == "__main__":
    test_rsp()



from __future__ import annotations

from typing import Generator, List, Optional, Tuple

import netsquid as ns
from netsquid.components.cchannel import ClassicalChannel
from netsquid.nodes import Node
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.message import Message
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostcomp import HostComponent
from qoala.sim.host.hostinterface import HostInterface
from qoala.util.tests import yield_from


class MockHostInterface(HostInterface):
    def __init__(self) -> None:
        self.remote: Optional[MockHostInterface] = None
        self.messages: List[Message] = []

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        assert self.remote is not None
        self.remote.messages.append(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        return self.messages.pop()
        yield  # to make it behave as a generator


def setup_alice_bob() -> Tuple[ClassicalSocket, ClassicalSocket]:
    alice = MockHostInterface()
    bob = MockHostInterface()
    alice.remote = bob
    bob.remote = alice

    return ClassicalSocket(alice, "bob", 0, 0), ClassicalSocket(bob, "alice", 0, 0)


def test_send_str():
    alice, bob = setup_alice_bob()

    alice.send_str("hello")
    msg = yield_from(bob.recv())
    assert msg == Message(0, 0, "hello")


def test_send_int():
    alice, bob = setup_alice_bob()

    alice.send_int(3)
    msg = yield_from(bob.recv_int())
    assert msg == 3


def test_send_float():
    alice, bob = setup_alice_bob()

    alice.send_float(3.14)
    msg = yield_from(bob.recv_float())
    assert msg == 3.14


def test_multiple_pids():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = HostComponent(alice, ehi_network)
    bob_comp = HostComponent(bob, ehi_network)

    channel_ab = ClassicalChannel("chan_ab", delay=1000)
    channel_ba = ClassicalChannel("chan_ba", delay=1000)

    alice_comp.peer_out_port("bob").connect(channel_ab.ports["send"])
    channel_ab.ports["recv"].connect(bob_comp.peer_in_port("alice"))

    bob_comp.peer_out_port("alice").connect(channel_ba.ports["send"])
    channel_ba.ports["recv"].connect(alice_comp.peer_in_port("bob"))

    alice_intf = HostInterface(alice_comp, ehi_network)
    bob_intf = HostInterface(bob_comp, ehi_network)

    alice_csck00 = ClassicalSocket(alice_intf, "bob", pid=0, remote_pid=0)
    alice_csck11 = ClassicalSocket(alice_intf, "bob", pid=1, remote_pid=1)
    bob_csck00 = ClassicalSocket(bob_intf, "alice", pid=0, remote_pid=0)
    bob_csck11 = ClassicalSocket(bob_intf, "alice", pid=1, remote_pid=1)

    class Alice(Protocol):
        def run(self) -> Generator[EventExpression, None, None]:
            alice_csck11.send_int(42)
            alice_csck00.send_int(3)

    class Bob(Protocol):
        def run(self) -> Generator[EventExpression, None, None]:
            assert bob_intf.get_available_messages("alice") == []
            yield from bob_intf.wait_for_msg("alice")
            assert (0, 0) in bob_intf.get_available_messages("alice")
            assert (1, 1) in bob_intf.get_available_messages("alice")
            value00 = bob_csck00.read_int()
            print(value00)
            assert value00 == 3
            value11 = bob_csck11.read_int()
            print(value11)
            assert value11 == 42

    alice_intf.start()
    bob_intf.start()
    Alice().start()
    Bob().start()

    ns.sim_run()


if __name__ == "__main__":
    test_send_str()
    test_send_int()
    test_send_float()
    test_multiple_pids()



import json
import os
from argparse import ArgumentParser
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import dacite
import matplotlib.pyplot as plt


@dataclass
class DataPoint:
    cc_latency: float
    num_qubits_bob: int
    tel_succ_prob: float
    tel_succ_prob_lower: float
    tel_succ_prob_upper: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int
    latency_factor: float
    net_bin_factors: List[float]
    prio_epr: bool
    num_qubits: int
    num_steps: int
    t1: float
    t2: float


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_png(filename: str):
    output_dir = relative_to_cwd("plots")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_path = os.path.join(output_dir, f"{filename}.png")
    plt.savefig(output_path)
    print(f"plot written to {output_path}")


def create_meta(filename: str, datas: Dict[str, Data]):
    output_dir = relative_to_cwd("plots")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_path = os.path.join(output_dir, f"{filename}.json")
    meta = {}
    meta["datafiles"] = [
        {
            "filename": datafile,
            "timestamp": data.meta.timestamp,
            "num_qubits": data.meta.num_qubits,
            "num_steps": data.meta.num_steps,
        }
        for datafile, data in datas.items()
    ]
    with open(output_path, "w") as metafile:
        json.dump(meta, metafile)


def load_data(path: str) -> Data:
    with open(relative_to_cwd(path), "r") as f:
        all_data = json.load(f)

    return dacite.from_dict(Data, all_data)


FORMATS = {
    1: "^-",
    2: "s-",
    5: "D-",
}

COLORS = {
    1: "#1F77B4",
    2: "#2CA02C",
    5: "#FF7F0E",
}


def plot_sweep_net_bin_period(timestamp: str, datas: Dict[str, Data]) -> None:
    fig, ax = plt.subplots()

    ax.grid()
    ax.set_xlabel(
        "Time slot length as fraction of node-node communication latency", fontsize=13
    )
    ax.set_ylabel("Makespan (ms)", fontsize=14)

    ax.tick_params(labelsize=12)

    fmts = ["o-b", "o-r", "o-k", "o-g", "s-y"]
    labels = [f"{data.meta.num_qubits} qubits" for data in datas.values()]

    nbf = [nbf for nbf in list(datas.values())[0].meta.net_bin_factors]
    for data, fmt, label in zip(datas.values(), fmts, labels):
        makespans = [p.makespan / 1e6 for p in data.data_points]
        ax.errorbar(x=nbf, y=makespans, fmt=FORMATS[data.meta.num_qubits], label=label)

    # ax.set_title(
    #     "Teleportation makespan vs time bin length in network schedule",
    #     wrap=True,
    # )

    ax.set_ylim(19, 105)
    ax.legend(loc="upper left", fontsize=11)

    create_png("LAST")
    create_png(timestamp)


def sweep_net_bin_factor_per_qubit_num(num_qubits: List[int], num_steps: int):
    datas: Dict[str, Data] = {}

    for nq in num_qubits:
        dir = f"data/sweep_bin_length_{nq}_{num_steps}"
        data = load_data(f"{dir}/LAST.json")
        data_timestamp = data.meta.timestamp
        datas[f"{dir}/{data_timestamp}.json"] = data

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    create_meta("LAST_meta", datas)
    create_meta(f"{timestamp}_meta", datas)
    plot_sweep_net_bin_period(timestamp, datas)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_qubits", "-q", type=int, nargs="+", required=True)
    parser.add_argument("--num_steps", "-s", type=int, required=True)

    args = parser.parse_args()

    num_qubits = args.num_qubits
    num_steps = args.num_steps

    sweep_net_bin_factor_per_qubit_num(num_qubits, num_steps)
